from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestComboSwitches(MpfTestCase):

    def get_config_file(self):
        return 'combo_switches.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/combo_switches/'

    def test_tag_combo(self):
        self.mock_event('tag_combo_both')
        self.mock_event('tag_combo_inactive')
        self.mock_event('tag_combo_one')

        self.hit_switch_and_run('switch5', 1)
        self.assertEventNotCalled('tag_combo_both')
        self.assertEventNotCalled('tag_combo_inactive')
        self.assertEventNotCalled('tag_combo_one')
        self.assertEqual(self.machine.combo_switches["tag_combo"].state, 'inactive')

        self.hit_switch_and_run('switch7', 1)
        self.assertEventCalled('tag_combo_both')
        self.assertEventNotCalled('tag_combo_inactive')
        self.assertEventNotCalled('tag_combo_one')
        self.assertEqual(self.machine.combo_switches["tag_combo"].state, 'both')

        self.release_switch_and_run('switch7', 1)
        self.assertEventNotCalled('tag_combo_inactive')
        self.assertEventCalled('tag_combo_one')
        self.assertEqual(self.machine.combo_switches["tag_combo"].state, 'one')

        self.mock_event('tag_combo_both')
        self.mock_event('tag_combo_inactive')
        self.mock_event('tag_combo_one')

        self.hit_switch_and_run('switch7', 1)
        self.assertEventCalled('tag_combo_both')
        self.assertEventNotCalled('tag_combo_inactive')
        self.assertEventNotCalled('tag_combo_one')

        self.release_switch_and_run('switch5', 1)
        self.assertEventNotCalled('tag_combo_inactive')
        self.assertEventCalled('tag_combo_one')

        self.release_switch_and_run('switch7', 1)
        self.assertEventCalled('tag_combo_inactive')

        # now make sure it all works with the switches in the other order
        self.mock_event('tag_combo_both')
        self.mock_event('tag_combo_inactive')
        self.mock_event('tag_combo_one')

        self.hit_switch_and_run('switch7', 1)
        self.assertEventNotCalled('tag_combo_both')
        self.assertEventNotCalled('tag_combo_inactive')
        self.assertEventNotCalled('tag_combo_one')

        self.hit_switch_and_run('switch5', 1)
        self.assertEventCalled('tag_combo_both')
        self.assertEventNotCalled('tag_combo_inactive')
        self.assertEventNotCalled('tag_combo_one')

        self.release_switch_and_run('switch5', 1)
        self.assertEventNotCalled('tag_combo_inactive')
        self.assertEventCalled('tag_combo_one')

        self.mock_event('tag_combo_both')
        self.mock_event('tag_combo_inactive')
        self.mock_event('tag_combo_one')

        self.hit_switch_and_run('switch5', 1)
        self.assertEventCalled('tag_combo_both')
        self.assertEventNotCalled('tag_combo_inactive')
        self.assertEventNotCalled('tag_combo_one')

        self.release_switch_and_run('switch7', 1)
        self.assertEventNotCalled('tag_combo_inactive')
        self.assertEventCalled('tag_combo_one')

        self.release_switch_and_run('switch5', 1)
        self.assertEventCalled('tag_combo_inactive')

    def test_switch_combo(self):
        self.mock_event('switch_combo_both')
        self.mock_event('switch_combo_inactive')
        self.mock_event('switch_combo_one')

        self.hit_switch_and_run('switch1', 1)
        self.assertEventNotCalled('switch_combo_both')
        self.assertEventNotCalled('switch_combo_inactive')
        self.assertEventNotCalled('switch_combo_one')

        self.hit_switch_and_run('switch2', 1)
        self.assertEventCalled('switch_combo_both')
        self.assertEventNotCalled('switch_combo_inactive')
        self.assertEventNotCalled('switch_combo_one')

        self.release_switch_and_run('switch2', 1)
        self.assertEventNotCalled('switch_combo_inactive')
        self.assertEventCalled('switch_combo_one')

        self.mock_event('switch_combo_both')
        self.mock_event('switch_combo_inactive')
        self.mock_event('switch_combo_one')

        self.hit_switch_and_run('switch2', 1)
        self.assertEventCalled('switch_combo_both')
        self.assertEventNotCalled('switch_combo_inactive')
        self.assertEventNotCalled('switch_combo_one')

        self.release_switch_and_run('switch1', 1)
        self.assertEventNotCalled('switch_combo_inactive')
        self.assertEventCalled('switch_combo_one')

        self.release_switch_and_run('switch2', 1)
        self.assertEventCalled('switch_combo_inactive')

        # test long offset time
        self.mock_event('switch_combo_both')
        self.mock_event('switch_combo_inactive')
        self.mock_event('switch_combo_one')

        self.hit_switch_and_run('switch1', 100)
        self.assertEventNotCalled('switch_combo_both')

        self.hit_switch_and_run('switch2', .1)
        self.assertEventCalled('switch_combo_both')

    def test_multiple_switch_combo(self):

        # first test the basics with multiple switches listed

        self.mock_event('multiple_switch_combo_both')
        self.mock_event('multiple_switch_combo_inactive')
        self.mock_event('multiple_switch_combo_one')

        self.hit_switch_and_run('switch1', 1)
        self.assertEventNotCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventNotCalled('multiple_switch_combo_one')

        self.hit_switch_and_run('switch3', 1)
        self.assertEventCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventNotCalled('multiple_switch_combo_one')

        self.release_switch_and_run('switch3', 1)
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventCalled('multiple_switch_combo_one')

        self.mock_event('multiple_switch_combo_both')
        self.mock_event('multiple_switch_combo_inactive')
        self.mock_event('multiple_switch_combo_one')

        self.hit_switch_and_run('switch3', 1)
        self.assertEventCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventNotCalled('multiple_switch_combo_one')

        self.release_switch_and_run('switch1', 1)
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventCalled('multiple_switch_combo_one')

        self.release_switch_and_run('switch3', 1)
        self.assertEventCalled('multiple_switch_combo_inactive')

        # now start playing with combinations of switches from the same group

        self.mock_event('multiple_switch_combo_both')
        self.mock_event('multiple_switch_combo_inactive')
        self.mock_event('multiple_switch_combo_one')

        # hit switch 1, nothing happens
        self.hit_switch_and_run('switch1', 1)
        self.assertEventNotCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventNotCalled('multiple_switch_combo_one')

        # hit switch 2, which is in group 1, so still nothing happens
        self.hit_switch_and_run('switch2', 1)
        self.assertEventNotCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventNotCalled('multiple_switch_combo_one')

        # hit switch 3, which is in group 2, so we're active
        self.hit_switch_and_run('switch3', 1)
        self.assertEventCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventNotCalled('multiple_switch_combo_one')

        self.mock_event('multiple_switch_combo_both')
        self.mock_event('multiple_switch_combo_inactive')
        self.mock_event('multiple_switch_combo_one')

        # hit switch 4, in group 2, so nothing happens
        self.hit_switch_and_run('switch4', 1)
        self.assertEventNotCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventNotCalled('multiple_switch_combo_one')

        # release switch 3, but switch 4 is still active, so nothing happens
        self.release_switch_and_run('switch3', 1)
        self.assertEventNotCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventNotCalled('multiple_switch_combo_one')

        # release switch 2, but switch 1 is still active, so nothing happens
        self.release_switch_and_run('switch2', 1)
        self.assertEventNotCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventNotCalled('multiple_switch_combo_one')

        # release switch 1, the last from group 1, so now we have the one event
        self.release_switch_and_run('switch1', 1)
        self.assertEventNotCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventCalled('multiple_switch_combo_one')

        # hit switch 2, so now we go back to combo active
        self.hit_switch_and_run('switch2', 1)
        self.assertEventCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventCalled('multiple_switch_combo_one')

        self.mock_event('multiple_switch_combo_both')
        self.mock_event('multiple_switch_combo_inactive')
        self.mock_event('multiple_switch_combo_one')

        # release switch 4, so we go back to one is active
        self.release_switch_and_run('switch4', 1)
        self.assertEventNotCalled('multiple_switch_combo_both')
        self.assertEventNotCalled('multiple_switch_combo_inactive')
        self.assertEventCalled('multiple_switch_combo_one')

        # release switch 2, so back to none are active
        self.release_switch_and_run('switch2', 1)
        self.assertEventNotCalled('multiple_switch_combo_both')
        self.assertEventCalled('multiple_switch_combo_inactive')
        self.assertEventCalled('multiple_switch_combo_one')

    def test_custom_offset(self):

        # we have a 1s offset

        self.mock_event('custom_offset_both')
        self.mock_event('custom_offset_inactive')
        self.mock_event('custom_offset_one')
        self.mock_event('custom_offset_switches_1')
        self.mock_event('custom_offset_switches_2')

        self.hit_switch_and_run('switch1', .1)
        self.assertEventNotCalled('custom_offset_switches_1')
        self.assertEventNotCalled('custom_offset_switches_2')
        self.advance_time_and_run(1.9)
        self.assertEventNotCalled('custom_offset_both')
        self.assertEventNotCalled('custom_offset_inactive')
        self.assertEventNotCalled('custom_offset_one')
        self.assertEventCalled('custom_offset_switches_1')
        self.assertEventNotCalled('custom_offset_switches_2')
        self.mock_event('custom_offset_switches_1')

        # switch2 in more than 1s offset time, so events should not be posted
        self.hit_switch_and_run('switch2', 1)
        self.assertEventNotCalled('custom_offset_both')
        self.assertEventNotCalled('custom_offset_inactive')
        self.assertEventNotCalled('custom_offset_one')
        self.assertEventNotCalled('custom_offset_switches_1')
        self.assertEventNotCalled('custom_offset_switches_2')

        self.release_switch_and_run('switch1', .1)
        self.release_switch_and_run('switch2', .1)

        # now hit both of the switches in < 1s
        self.hit_switch_and_run('switch1', .1)
        self.hit_switch_and_run('switch2', .1)
        self.assertEventCalled('custom_offset_both')
        self.advance_time_and_run(10)
        self.assertEventNotCalled('custom_offset_inactive')
        self.assertEventNotCalled('custom_offset_one')
        self.assertEventNotCalled('custom_offset_switches_1')
        self.assertEventNotCalled('custom_offset_switches_2')

    def test_custom_hold(self):

        # we have a 1s hold time

        self.mock_event('custom_hold_both')
        self.mock_event('custom_hold_inactive')
        self.mock_event('custom_hold_one')

        self.hit_switch_and_run('switch1', 5)
        self.hit_switch_and_run('switch2', .5)
        self.assertEventNotCalled('custom_hold_both')
        self.assertEventNotCalled('custom_hold_inactive')
        self.assertEventNotCalled('custom_hold_one')

        # advance more than 1s from the first switch, nothing should happen
        self.advance_time_and_run(.1)
        self.assertEventNotCalled('custom_hold_both')
        self.assertEventNotCalled('custom_hold_inactive')
        self.assertEventNotCalled('custom_hold_one')

        # advance more than 1s from the second switch
        self.advance_time_and_run(.6)
        self.assertEventCalled('custom_hold_both')
        self.assertEventNotCalled('custom_hold_inactive')
        self.assertEventNotCalled('custom_hold_one')

        # release one of the switches, one should be posted
        self.release_switch_and_run('switch2', .1)
        self.assertEventNotCalled('custom_hold_inactive')
        self.assertEventCalled('custom_hold_one')

        self.mock_event('custom_hold_both')
        self.mock_event('custom_hold_inactive')
        self.mock_event('custom_hold_one')

        # hit the second switch again, event should not be posted
        self.hit_switch_and_run('switch2', .5)
        self.assertEventNotCalled('custom_hold_both')
        self.assertEventNotCalled('custom_hold_inactive')
        self.assertEventNotCalled('custom_hold_one')

        # advance more than 1s to see the event
        self.advance_time_and_run(.6)
        self.assertEventCalled('custom_hold_both')
        self.assertEventNotCalled('custom_hold_inactive')
        self.assertEventNotCalled('custom_hold_one')

        # release both
        self.release_switch_and_run('switch1', .1)
        self.release_switch_and_run('switch2', .1)
        self.assertEventCalled('custom_hold_inactive')

    def test_custom_release(self):

        # release time of 1s

        self.mock_event('custom_release_both')
        self.mock_event('custom_release_inactive')
        self.mock_event('custom_release_one')

        # both switches should post both
        self.hit_switch_and_run('switch1', 1)
        self.hit_switch_and_run('switch2', .1)
        self.assertEventCalled('custom_release_both')
        self.assertEventNotCalled('custom_release_inactive')
        self.assertEventNotCalled('custom_release_one')

        # release 1, should not post one because it's less than 1s
        self.release_switch_and_run('switch2', .1)
        self.assertEventNotCalled('custom_release_inactive')
        self.assertEventNotCalled('custom_release_one')

        # wait more than 1s and one should be posted
        self.advance_time_and_run(1)
        self.assertEventCalled('custom_release_one')

        # release the other switch, inactive should not be posted yet
        self.release_switch_and_run('switch1', .1)
        self.assertEventNotCalled('custom_release_inactive')

        # wait more than 1s for the inactive event
        self.advance_time_and_run(1)
        self.assertEventCalled('custom_release_inactive')

        # start over
        self.hit_switch_and_run('switch1', 1)
        self.hit_switch_and_run('switch2', 1)

        self.mock_event('custom_release_both')
        self.mock_event('custom_release_inactive')
        self.mock_event('custom_release_one')

        # release and reactivate in less than 1s, no new events
        self.release_switch_and_run('switch2', .5)
        self.hit_switch_and_run('switch2', .1)

        self.assertEventNotCalled('custom_release_both')
        self.assertEventNotCalled('custom_release_inactive')
        self.assertEventNotCalled('custom_release_one')

        # make sure no new events after the initial release time passed
        self.advance_time_and_run(1)
        self.assertEventNotCalled('custom_release_both')
        self.assertEventNotCalled('custom_release_inactive')
        self.assertEventNotCalled('custom_release_one')

        # release and reactive both in less than 1s, no new events
        self.release_switch_and_run('switch1', .1)
        self.release_switch_and_run('switch2', .1)
        self.hit_switch_and_run('switch1', .1)
        self.hit_switch_and_run('switch2', .1)

        self.assertEventNotCalled('custom_release_both')
        self.assertEventNotCalled('custom_release_inactive')
        self.assertEventNotCalled('custom_release_one')

        # make sure no new events
        self.advance_time_and_run(1)
        self.assertEventNotCalled('custom_release_both')
        self.assertEventNotCalled('custom_release_inactive')
        self.assertEventNotCalled('custom_release_one')

        # now do the whole thing again, with the switches flipped

        self.release_switch_and_run('switch2', 2)

        self.mock_event('custom_release_both')
        self.mock_event('custom_release_inactive')
        self.mock_event('custom_release_one')

        # both switches should post both
        self.hit_switch_and_run('switch2', 1)
        self.hit_switch_and_run('switch1', .1)
        self.assertEventCalled('custom_release_both')
        self.assertEventNotCalled('custom_release_inactive')
        self.assertEventNotCalled('custom_release_one')

        # release 1, should not post one because it's less than 1s
        self.release_switch_and_run('switch1', .1)
        self.assertEventNotCalled('custom_release_inactive')
        self.assertEventNotCalled('custom_release_one')

        # wait more than 1s and one should be posted
        self.advance_time_and_run(1)
        self.assertEventCalled('custom_release_one')

        # release the other switch, inactive should not be posted yet
        self.release_switch_and_run('switch2', .1)
        self.assertEventNotCalled('custom_release_inactive')

        # wait more than 1s for the inactive event
        self.advance_time_and_run(1)
        self.assertEventCalled('custom_release_inactive')

        # start over
        self.hit_switch_and_run('switch2', 1)
        self.hit_switch_and_run('switch1', .1)

        self.mock_event('custom_release_both')
        self.mock_event('custom_release_inactive')
        self.mock_event('custom_release_one')

        # release and reactivate in less than 1s, no new events
        self.release_switch_and_run('switch1', .5)
        self.hit_switch_and_run('switch1', .1)

        self.assertEventNotCalled('custom_release_both')
        self.assertEventNotCalled('custom_release_inactive')
        self.assertEventNotCalled('custom_release_one')

    def test_custom_times_multiple_switches(self):

        # this is a sort of catch all with all three types of custom times,
        # but with multiple switches

        # time is >1s from the first switch
        self._reset_custom_times_multiple_switches()
        self.hit_switch_and_run('switch5', .5)
        self.hit_switch_and_run('switch6', .6)
        self.hit_switch_and_run('switch7', .1)
        self.assertEventNotCalled('custom_times_multiple_switches_both')

        # time is <1s from first switch
        self._reset_custom_times_multiple_switches()
        self.hit_switch_and_run('switch5', .5)
        self.hit_switch_and_run('switch6', .1)
        self.hit_switch_and_run('switch7', .1)
        self.assertEventNotCalled('custom_times_multiple_switches_both')

        # there's a 1s hold time
        self.advance_time_and_run(1.1)
        self.assertEventCalled('custom_times_multiple_switches_both')

        # release switch7, one event should post after 1s
        self.release_switch_and_run('switch7', .1)
        self.assertEventNotCalled('custom_times_multiple_switches_one')
        self.advance_time_and_run(1)
        self.assertEventCalled('custom_times_multiple_switches_one')

        # test hold time
        self._reset_custom_times_multiple_switches()
        self.hit_switch_and_run('switch5', .1)
        self.hit_switch_and_run('switch6', .1)
        self.hit_switch_and_run('switch7', .1)
        self.assertEventNotCalled('custom_times_multiple_switches_both')
        self.advance_time_and_run(1)
        self.assertEventCalled('custom_times_multiple_switches_both')

    def _reset_custom_times_multiple_switches(self):
        self.release_switch_and_run('switch5', .1)
        self.release_switch_and_run('switch6', .1)
        self.release_switch_and_run('switch7', .1)
        self.release_switch_and_run('switch8', .1)
        self.advance_time_and_run(2)
        self.mock_event('custom_times_multiple_switches_both')
        self.mock_event('custom_times_multiple_switches_inactive')
        self.mock_event('custom_times_multiple_switches_one')

    def test_custom_events(self):
        self.mock_event('custom_events_both')
        self.mock_event('custom_events_inactive')
        self.mock_event('custom_events_one')
        self.mock_event('active_event')
        self.mock_event('active_event2')
        self.mock_event('inactive_event')
        self.mock_event('one_event')

        self.hit_switch_and_run('switch1', .1)
        self.hit_switch_and_run('switch2', .1)
        self.release_switch_and_run('switch1', .1)
        self.release_switch_and_run('switch2', .1)

        self.assertEventNotCalled('custom_events_both')
        self.assertEventNotCalled('custom_events_inactive')
        self.assertEventNotCalled('custom_events_one')
        self.assertEventCalled('active_event')
        self.assertEventCalled('active_event2')
        self.assertEventCalled('inactive_event')
        self.assertEventCalled('one_event')

    def test_combo_switches_in_mode(self):
        self.mock_event('mode1_combo_both')
        self.mock_event('mode1_combo_inactive')
        self.mock_event('mode1_combo_one')

        self.hit_switch_and_run('switch1', .1)
        self.hit_switch_and_run('switch2', .1)
        self.release_switch_and_run('switch1', .1)
        self.release_switch_and_run('switch2', .1)

        self.assertEventNotCalled('mode1_combo_both')
        self.assertEventNotCalled('mode1_combo_inactive')
        self.assertEventNotCalled('mode1_combo_one')

        self.advance_time_and_run(5)

        self.machine.modes["mode1"].start()
        self.advance_time_and_run()

        self.hit_switch_and_run('switch1', .1)
        self.hit_switch_and_run('switch2', .1)
        self.release_switch_and_run('switch1', .1)
        self.release_switch_and_run('switch2', .1)

        self.assertEventCalled('mode1_combo_both')
        self.assertEventCalled('mode1_combo_inactive')
        self.assertEventCalled('mode1_combo_one')

        self.machine.modes["mode1"].stop()
        self.advance_time_and_run()
        self.mock_event('mode1_combo_both')
        self.mock_event('mode1_combo_inactive')
        self.mock_event('mode1_combo_one')

        self.hit_switch_and_run('switch1', .1)
        self.hit_switch_and_run('switch2', .1)
        self.release_switch_and_run('switch1', .1)
        self.release_switch_and_run('switch2', .1)

        self.assertEventNotCalled('mode1_combo_both')
        self.assertEventNotCalled('mode1_combo_inactive')
        self.assertEventNotCalled('mode1_combo_one')

    def test_built_in_combos(self):
        self.mock_event('flipper_cancel')
        self.hit_switch_and_run('switch9', .1)
        self.hit_switch_and_run('switch10', .1)
        self.assertEventCalled('flipper_cancel')

        # make sure it works with long times too
        self.release_switch_and_run('switch9', .1)
        self.release_switch_and_run('switch10', .1)

        self.mock_event('flipper_cancel')
        self.hit_switch_and_run('switch9', 10)
        self.hit_switch_and_run('switch10', .1)
        self.assertEventCalled('flipper_cancel')



"""Contains the base class for ball devices."""
import asyncio
from collections import deque

from mpf.core.events import QueuedEvent, event_handler
from mpf.devices.ball_device.ball_count_handler import BallCountHandler
from mpf.devices.ball_device.ball_device_ejector import BallDeviceEjector

from mpf.core.delays import DelayManager
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.system_wide_device import SystemWideDevice
from mpf.core.utility_functions import Util


from mpf.devices.ball_device.incoming_balls_handler import IncomingBallsHandler, IncomingBall
from mpf.devices.ball_device.outgoing_balls_handler import OutgoingBallsHandler, OutgoingBall


@DeviceMonitor("available_balls", _state="state", counted_balls="balls")
class BallDevice(SystemWideDevice):

    """Base class for a 'Ball Device' in a pinball machine.

    A ball device is anything that can hold one or more balls, such as a
    trough, an eject hole, a VUK, a catapult, etc.

    Args: Same as Device.
    """

    config_section = 'ball_devices'
    collection = 'ball_devices'
    class_label = 'ball_device'

    __slots__ = ["delay", "available_balls", "_target_on_unexpected_ball", "_source_devices", "_ball_requests",
                 "ejector", "ball_count_handler", "incoming_balls_handler", "outgoing_balls_handler",
                 "counted_balls", "_state"]

    def __init__(self, machine, name):
        """Initialise ball device."""
        super().__init__(machine, name)

        self.delay = DelayManager(machine)

        self.available_balls = 0
        """Number of balls that are available to be ejected. This differs from
        `balls` since it's possible that this device could have balls that are
        being used for some other eject, and thus not available."""

        self._target_on_unexpected_ball = None
        # Device will eject to this target when it captures an unexpected ball

        self._source_devices = list()
        # Ball devices that have this device listed among their eject targets

        self._ball_requests = deque()
        # deque of tuples that holds requests from target devices for balls
        # that this device could fulfil
        # each tuple is (target device, boolean player_controlled flag)

        self.ejector = None                     # type: BallDeviceEjector
        self.ball_count_handler = None          # type: BallCountHandler
        self.incoming_balls_handler = None      # type: IncomingBallsHandler
        self.outgoing_balls_handler = None      # type: OutgoingBallsHandler

        # mirrored from ball_count_handler to make it obserable by the monitor
        self.counted_balls = 0
        self._state = "idle"

    def set_eject_state(self, state):
        """Set the current device state."""
        self.info_log("State: %s", state)
        self._state = state

    @property
    def balls(self):
        """Return the number of balls we expect in the near future."""
        if self._state in ["ball_left", "failed_confirm"]:
            return self.counted_balls - 1
        return self.counted_balls

    @event_handler(11)
    def event_entrance(self, **kwargs):
        """Event handler for entrance events."""
        del kwargs
        self.ball_count_handler.counter.received_entrance_event()

    async def _initialize(self):
        """Initialize right away."""
        await super()._initialize()
        self._configure_targets()

        self.ball_count_handler = BallCountHandler(self)
        self.incoming_balls_handler = IncomingBallsHandler(self)
        self.outgoing_balls_handler = OutgoingBallsHandler(self)

        # delay ball counters because we have to wait for switches to be ready
        self.machine.events.add_handler('init_phase_2', self._initialize_late)

    def _initialize_late(self, queue: QueuedEvent, **kwargs):
        """Create ball counters."""
        del kwargs
        queue.wait()
        complete_future = asyncio.ensure_future(self._initialize_async())
        complete_future.add_done_callback(lambda x: queue.clear())

    def stop_device(self):
        """Stop device."""
        self.debug_log("Stopping ball device")
        if self.ball_count_handler:
            self.ball_count_handler.stop()
            self.incoming_balls_handler.stop()
            self.outgoing_balls_handler.stop()

    async def expected_ball_received(self):
        """Handle an expected ball."""
        # post enter event
        unclaimed_balls = await self._post_enter_event(unclaimed_balls=0, new_available_balls=0)
        # there might still be unclaimed balls (e.g. because of a ball_routing)
        self._balls_added_callback(0, unclaimed_balls)

    async def unexpected_ball_received(self):
        """Handle an unexpected ball."""
        # capture from playfield
        await self._post_capture_from_playfield_event()
        # post enter event
        unclaimed_balls = await self._post_enter_event(unclaimed_balls=1, new_available_balls=1)
        # add available_balls and route unclaimed ball to the default target
        self._balls_added_callback(1, unclaimed_balls)

    async def handle_mechanial_eject_during_idle(self):
        """Handle mechanical eject."""
        # handle lost balls via outgoing balls handler (if mechanical eject)
        self.config['eject_targets'][0].available_balls += 1
        eject = OutgoingBall(self.config['eject_targets'][0])
        eject.eject_timeout = self.config['eject_timeouts'][eject.target] / 1000
        eject.max_tries = self.config['max_eject_attempts']
        eject.player_controlled = True
        eject.already_left = True
        self.outgoing_balls_handler.add_eject_to_queue(eject)

    async def lost_idle_ball(self):
        """Lost an ball while the device was idle."""
        # handle lost balls
        self.warning_log("Ball disappeared while idle. This should not normally happen.")
        self.available_balls -= 1
        self.config['ball_missing_target'].add_missing_balls(1)
        await self._balls_missing(1)

    async def lost_ejected_ball(self, target):
        """Handle an outgoing lost ball."""
        # follow path and check if we should request a new ball to the target or cancel the path
        if target.is_playfield():
            raise AssertionError("Lost a ball to playfield {}. This should not happen".format(target))
        if target.cancel_path_if_target_is(self, self.config['ball_missing_target']):
            # add ball to default target because it would have gone there anyway
            self.warning_log("Path to %s canceled. Assuming the ball jumped to %s.", target,
                             self.config['ball_missing_target'])
        elif target.find_available_ball_in_path(self):
            self.warning_log("Path is not going to ball_missing_target %s. Restoring path by requesting new ball to "
                             "target %s.", self.config['ball_missing_target'], target)
            # remove one ball first because it will get a new one with the eject
            target.available_balls -= 1
            self.eject(target=target)
        else:
            self.warning_log("Failed to restore the path. If you can reproduce this please report in the forum!")

        self.config['ball_missing_target'].add_missing_balls(1)
        await self._balls_missing(1)

    async def lost_incoming_ball(self, source):
        """Handle lost ball which was confirmed to have left source."""
        del source
        if self.cancel_path_if_target_is(self, self.config['ball_missing_target']):
            # add ball to default target
            self.warning_log("Path to canceled. Assuming the ball jumped to %s.", self.config['ball_missing_target'])
        elif self.find_available_ball_in_path(self):
            self.warning_log("Path is not going to ball_missing_target %s. Restoring path by requesting a new ball.",
                             self.config['ball_missing_target'])
            self.available_balls -= 1
            self.request_ball()
        else:
            self.warning_log("Failed to restore the path. If you can reproduce this please report in the forum!")

        self.config['ball_missing_target'].add_missing_balls(1)
        await self._balls_missing(1)

    def cancel_path_if_target_is(self, start, target):
        """Check if the ball is going to a certain target and cancel the path in that case."""
        return self.outgoing_balls_handler.cancel_path_if_target_is(start, target)

    def find_available_ball_in_path(self, start):
        """Try to remove available ball at the end of the path."""
        return self.outgoing_balls_handler.find_available_ball_in_path(start)

    async def _initialize_async(self):
        """Count balls without handling them as new."""
        await self.ball_count_handler.initialise()
        await self.incoming_balls_handler.initialise()
        await self.outgoing_balls_handler.initialise()

        self.available_balls = self.ball_count_handler.handled_balls

    async def _post_capture_from_playfield_event(self):
        await self.machine.events.post_async('balldevice_captured_from_{}'.format(
            self.config['captures_from'].name),
            balls=1)
        '''event: balldevice_captured_from_(captures_from)

        desc: A ball device has just captured a ball from the device called
        (captures_from)

        args:
        balls: The number of balls that were captured.

        '''

    async def _post_enter_event(self, unclaimed_balls, new_available_balls):
        self.debug_log("Processing new ball")
        result = await self.machine.events.post_relay_async('balldevice_{}_ball_enter'.format(
            self.name),
            new_balls=1,
            unclaimed_balls=unclaimed_balls,
            new_available_balls=new_available_balls,
            device=self)
        '''event: balldevice_(name)_ball_enter

        desc: A ball (or balls) have just entered the ball device called
        "name".

        Note that this is a relay event based on the "unclaimed_balls" arg. Any
        unclaimed balls in the relay will be processed as new balls entering
        this device.

        Please be aware that we did not add those balls to balls or available_balls of the device during this event.

        args:

        unclaimed_balls: The number of balls that have not yet been claimed.
        device: A reference to the ball device object that is posting this
        event.
        '''
        return result['unclaimed_balls']

    def add_incoming_ball(self, incoming_ball: IncomingBall):
        """Notify this device that there is a ball heading its way."""
        self.incoming_balls_handler.add_incoming_ball(incoming_ball)

    def remove_incoming_ball(self, incoming_ball: IncomingBall):
        """Remove a ball from the incoming balls queue."""
        self.incoming_balls_handler.remove_incoming_ball(incoming_ball)

    def wait_for_ready_to_receive(self, source):
        """Wait until this device is ready to receive a ball."""
        return self.ball_count_handler.wait_for_ready_to_receive(source)

    @property
    def requested_balls(self):
        """Return the number of requested balls."""
        return len(self._ball_requests)

    def _source_device_balls_available(self, **kwargs) -> None:
        del kwargs
        if self._ball_requests:
            (target, player_controlled) = self._ball_requests.popleft()
            self._setup_or_queue_eject_to_target(target, player_controlled)

    # ---------------------- End of state handling code -----------------------

    def _parse_config(self):
        # ensure eject timeouts list matches the length of the eject targets
        if (len(self.config['eject_timeouts']) <
                len(self.config['eject_targets'])):
            self.config['eject_timeouts'] += ["10s"] * (
                len(self.config['eject_targets']) -
                len(self.config['eject_timeouts']))

        if (len(self.config['ball_missing_timeouts']) <
                len(self.config['eject_targets'])):
            self.config['ball_missing_timeouts'] += ["20s"] * (
                len(self.config['eject_targets']) -
                len(self.config['ball_missing_timeouts']))

        timeouts_list = self.config['eject_timeouts']
        self.config['eject_timeouts'] = dict()

        for i in range(len(self.config['eject_targets'])):
            self.config['eject_timeouts'][self.config['eject_targets'][i]] = (
                Util.string_to_ms(timeouts_list[i]))

        timeouts_list = self.config['ball_missing_timeouts']
        self.config['ball_missing_timeouts'] = dict()

        for i in range(len(self.config['eject_targets'])):
            self.config['ball_missing_timeouts'][
                self.config['eject_targets'][i]] = (
                Util.string_to_ms(timeouts_list[i]))
        # End code to create timeouts list ------------------------------------

    @property
    def capacity(self):
        """Return the ball capacity."""
        return self.ball_count_handler.counter.capacity

    def _validate_config(self):
        # perform logical validation
        # a device cannot have hold_coil and eject_coil
        if (not self.config.get('eject_coil') and not self.config.get('hold_coil') and
                not self.config['mechanical_eject'] and not self.config.get('ejector', False)):
            self.raise_config_error('Configuration error in {} ball device. '
                                    'Device needs an eject_coil, a hold_coil, or '
                                    '"mechanical_eject: True"'.format(self.name), 4)

        # all eject_timeout < all ball_missing_timeouts
        if max(self.config['eject_timeouts'].values()) > min(
                self.config['ball_missing_timeouts'].values()):
            self.raise_config_error('Configuration error in {} ball device. '
                                    'all ball_missing_timeouts have to be larger '
                                    'than all eject_timeouts'.
                                    format(self.name), 8)

        # all ball_missing_timeouts < incoming ball timeout
        if max(self.config['ball_missing_timeouts'].values()) > 60000:
            self.raise_config_error('Configuration error in {} ball device. '
                                    'incoming ball timeout has to be larger '
                                    'than all ball_missing_timeouts'.
                                    format(self.name), 9)

        if (self.config['confirm_eject_type'] == "switch" and
                not self.config['confirm_eject_switch']):
            self.raise_config_error("When using confirm_eject_type switch you " +
                                    "to specify a confirm_eject_switch", 7)

        if (self.config['confirm_eject_type'] == "event" and
                not self.config['confirm_eject_event']):
            self.raise_config_error("When using confirm_eject_type event you " +
                                    "to specify a confirm_eject_event", 14)

        if "ball_add_live" in self.tags:
            self.raise_config_error("Using \"tag: ball_add_live\" is deprecated. Please use default_source_device "
                                    "in your playfield section instead.", 10)

        if "drain" in self.tags and "trough" not in self.tags and not self.find_next_trough():
            self.raise_config_error("No path to trough but device is tagged as drain", 11)

        if ("drain" not in self.tags and "trough" not in self.tags and
                not self.find_path_to_target(self._target_on_unexpected_ball)):
            self.raise_config_error("BallDevice {} has no path to target_on_unexpected_ball '{}'".format(
                self.name, self._target_on_unexpected_ball.name), 12)

    def load_config(self, config):
        """Load config."""
        super().load_config(config)

        # load targets and timeouts
        self._parse_config()

    def _configure_targets(self):
        if self.config['target_on_unexpected_ball']:
            self._target_on_unexpected_ball = self.config['target_on_unexpected_ball']
        else:
            self._target_on_unexpected_ball = self.config['captures_from']

        # validate that configuration is valid
        self._validate_config()

        ejector_config = self.config.get("ejector", {})

        # no ejector config. support legacy config
        if not ejector_config:
            if self.config.get('eject_coil'):
                if self.config.get('eject_coil_enable_time'):
                    ejector_config["class"] = "mpf.devices.ball_device.enable_coil_ejector.EnableCoilEjector"
                else:
                    ejector_config["class"] = "mpf.devices.ball_device.pulse_coil_ejector.PulseCoilEjector"
            elif self.config.get('hold_coil'):
                ejector_config["class"] = "mpf.devices.ball_device.hold_coil_ejector.HoldCoilEjector"

        if not ejector_config:
            self.debug_log("Device does not have any ejector.")
        else:
            ejector_class = Util.string_to_class(ejector_config["class"])
            if not ejector_class:
                self.raise_config_error("Could not load ejector {}".format(ejector_config["class"]), 1)

            self.ejector = ejector_class(ejector_config, self, self.machine)

        if self.ejector and self.config['ball_search_order']:
            self.config['captures_from'].ball_search.register(
                self.config['ball_search_order'], self.ejector.ball_search,
                self.name)

        # Register events to watch for ejects targeted at this device
        for device in self.machine.ball_devices.values():
            if device.is_playfield():
                continue
            for target in device.config['eject_targets']:
                if target.name == self.name:
                    self._source_devices.append(device)
                    break

        # register event handler for available balls at source devices
        self.machine.events.add_handler(
            'balldevice_balls_available',
            self._source_device_balls_available)

    def _balls_added_callback(self, new_balls, unclaimed_balls):
        # If we still have unclaimed_balls here, that means that no one claimed
        # them, so essentially they're "stuck." So we just eject them unless
        # this device is tagged 'trough' in which case we let it keep them.
        self.debug_log("Adding ball")
        self.available_balls += new_balls

        if unclaimed_balls and self.available_balls > 0:
            if 'trough' in self.tags:
                # ball already reached trough. everything is fine
                pass
            elif 'drain' in self.tags:
                # try to eject to next trough
                trough = self.find_next_trough()

                if not trough:
                    raise AssertionError("Could not find path to trough")

                for dummy_iterator in range(unclaimed_balls):
                    self._setup_or_queue_eject_to_target(trough)
            else:
                target = self._target_on_unexpected_ball

                # try to eject to configured target
                path = self.find_path_to_target(target)

                if not path:
                    raise AssertionError("Could not find path to playfield {}".format(target.name))

                self.info_log("Ejecting %s unexpected balls using path %s", unclaimed_balls, path)

                for dummy_iterator in range(unclaimed_balls):
                    self.setup_eject_chain(path, not self.config['auto_fire_on_unexpected_ball'])

        # we might have ball requests locally. serve them first
        if self._ball_requests:
            self._source_device_balls_available()

        # tell targets that we have balls available
        for dummy_iterator in range(new_balls):
            self.machine.events.post_boolean('balldevice_balls_available')

        self.machine.events.post('balldevice_{}_ball_entered'.format(self.name), new_balls=new_balls, device=self)
        '''event: balldevice_(name)_ball_entered

        desc: A ball (or balls) have just entered the ball device called
        "name".

        The ball was also added to balls and available_balls of the device.

        args:

        new_balls: The number of new balls that have not been claimed (by locks or similar).
        device: A reference to the ball device object that is posting this
        event.
        '''

    async def _balls_missing(self, balls):
        # Called when ball_count finds that balls are missing from this device
        self.info_log("%s ball(s) missing from device. Mechanical eject?"
                      " %s", abs(balls), self.config['mechanical_eject'])

        await self.machine.events.post_async('balldevice_{}_ball_missing'.format(self.name), balls=abs(balls))
        '''event: balldevice_(name)_ball_missing
        desc: The device (name) is missing a ball. Note this event is
        posted in addition to the generic *balldevice_ball_missing* event.
        args:
            balls: The number of balls that are missing
        '''
        await self.machine.events.post_async('balldevice_ball_missing', balls=abs(balls), name=self.name)
        '''event: balldevice_ball_missing
        desc: A ball is missing from a device.
        args:
            balls: The number of balls that are missing
            name: Name of device which lost the ball
        '''

    @property
    def state(self):
        """Return the device state."""
        return self._state

    def find_one_available_ball(self, path=None):
        """Find a path to a source device which has at least one available ball."""
        if path is None:
            path = deque()
        else:
            # copy path
            path = deque(path)

        # prevent loops
        if self in path:
            return False

        path.appendleft(self)

        if self.available_balls > 0 and len(path) > 1:
            return path

        for source in self._source_devices:
            full_path = source.find_one_available_ball(path=path)
            if full_path:
                return full_path

        return False

    @event_handler(1)
    def event_request_ball(self, balls=1, **kwargs):
        """Handle request_ball control event."""
        del kwargs
        self.request_ball(balls)

    def request_ball(self, balls=1):
        """Request that one or more balls is added to this device.

        Args:
        ----
            balls: Integer of the number of balls that should be added to this
                device. A value of -1 will cause this device to try to fill
                itself.
            **kwargs: unused
        """
        self.debug_log("Requesting Ball(s). Balls=%s", balls)

        for dummy_iterator in range(balls):
            self._setup_or_queue_eject_to_target(self)

        return balls

    def _setup_or_queue_eject_to_target(self, target, player_controlled=False):
        path_to_target = self.find_path_to_target(target)
        if target != self and not path_to_target:
            raise AssertionError("Do not know how to eject to {}".format(target.name))

        if self.available_balls > 0 and self != target:
            path = path_to_target
        else:

            path = self.find_one_available_ball()
            if not path:
                # put into queue here
                self._ball_requests.append((target, player_controlled))
                return False

            if target != self:
                path_to_target.popleft()    # remove self from path
                path.extend(path_to_target)

        path[0].setup_eject_chain(path, player_controlled)

        return True

    def setup_player_controlled_eject(self, target=None):
        """Set up a player controlled eject."""
        self.info_log("Setting up player-controlled eject. Balls: %s, "
                      "Target: %s, player_controlled_eject_event: %s",
                      1, target,
                      self.config['player_controlled_eject_event'])

        if self.config['mechanical_eject'] or (
                self.config['player_controlled_eject_event'] and self.ejector):

            self._setup_or_queue_eject_to_target(target, True)

        else:
            self.eject(target=target)

    def setup_eject_chain(self, path, player_controlled=False):
        """Set up an eject chain."""
        path = deque(path)
        if self.available_balls <= 0:
            raise AssertionError("Tried to setup an eject chain, but there are"
                                 " no available balls. Device: {}, Path: {}"
                                 .format(self.name, path))

        self.available_balls -= 1

        target = path[len(path) - 1]
        source = path.popleft()
        if source != self:
            raise AssertionError("Path starts somewhere else!")

        self.setup_eject_chain_next_hop(path, player_controlled)

        target.available_balls += 1

        self.machine.events.post_boolean('balldevice_balls_available')
        '''event: balldevice_balls_available
        desc: A device has balls available to be ejected.
        '''

    def setup_eject_chain_next_hop(self, path, player_controlled):
        """Set up one hop of the eject chain."""
        next_hop = path.popleft()
        self.debug_log("Adding eject chain")

        if next_hop not in self.config['eject_targets']:
            raise AssertionError("Broken path")

        eject = OutgoingBall(next_hop)
        eject.eject_timeout = self.config['eject_timeouts'][next_hop] / 1000
        eject.max_tries = self.config['max_eject_attempts']
        eject.player_controlled = player_controlled

        self.outgoing_balls_handler.add_eject_to_queue(eject)

        # check if we traversed the whole path
        if path:
            next_hop.setup_eject_chain_next_hop(path, player_controlled)

    def find_next_trough(self):
        """Find next trough after device."""
        # are we a trough?
        if 'trough' in self.tags:
            return self

        # otherwise find any target which can
        for target_device in self.config['eject_targets']:
            if target_device.is_playfield():
                continue
            trough = target_device.find_next_trough()
            if trough:
                return trough

        return False

    def find_path_to_target(self, target):
        """Find a path to this target."""
        # if we can eject to target directly just do it
        if target in self.config['eject_targets']:
            path = deque()
            path.appendleft(target)
            path.appendleft(self)
            return path

        # otherwise find any target which can
        for target_device in self.config['eject_targets']:
            if target_device.is_playfield():
                continue
            path = target_device.find_path_to_target(target)
            if path:
                path.appendleft(self)
                return path

        return False

    @event_handler(2)
    def event_eject(self, balls=1, target=None, **kwargs):
        """Handle eject control event."""
        del kwargs
        self.eject(balls, target)

    def eject(self, balls=1, target=None) -> int:
        """Eject balls to target.

        Return the number of balls found for eject. The remaining balls are queued for eject when available.
        """
        if not target:
            target = self._target_on_unexpected_ball

        self.info_log('Adding %s ball(s) to the eject_queue with target %s.',
                      balls, target)

        balls_found = 0
        # add request to queue
        for dummy_iterator in range(balls):
            if self._setup_or_queue_eject_to_target(target):
                balls_found += 1

        return balls_found

    @event_handler(3)
    def event_eject_all(self, target=None, **kwargs):
        """Handle eject_all control event."""
        del kwargs
        self.eject_all(target)

    def eject_all(self, target=None) -> bool:
        """Eject all the balls from this device.

        Args:
        ----
            target: The string or BallDevice target for this eject. Default of
                None means `playfield`.
            **kwargs: unused

        Returns True if there are balls to eject. False if this device is empty.
        """
        self.debug_log("Ejecting all balls")
        if self.available_balls > 0:
            self.eject(balls=self.available_balls, target=target)
            return True

        return False

    @classmethod
    def is_playfield(cls):
        """Return True if this ball device is a Playfield-type device, False if it's a regular ball device."""
        return False



"""Default ball search implementation for ejectors."""


class DefaultBallSearch:

    """Implement the default ball search behavior."""

    def ball_search(self, phase, iteration):
        """Run ball search."""
        del iteration
        if 'no-eject-on-ballsearch' in self.ball_device.config['tags']:
            return False
        if phase == 1:
            # round 1: only idle + no ball
            # only run ball search when the device is idle and contains no balls
            if self.ball_device.state == "idle" and self.ball_device.balls == 0:
                return self._fire_coil_for_search(True)
        elif phase == 2:
            # round 2: all devices except trough. small pulse
            if 'trough' not in self.ball_device.config['tags']:
                return self._fire_coil_for_search(False)
        else:
            # round 3: all devices except trough. normal pulse
            if 'trough' not in self.ball_device.config['tags']:
                return self._fire_coil_for_search(True)
        # no action by default
        return False

    def _fire_coil_for_search(self, full_power):
        """Fire eject coil for search."""
        raise NotImplementedError()



import time

from mpf.core.logging import LogMixin

from mpf.tests.MpfGameTestCase import MpfGameTestCase


class BenchmarkEvents(MpfGameTestCase):

    def get_options(self):
        options = super().get_options()
        if self.unittest_verbosity() <= 1:
            options["production"] = True
        return options

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'benchmarks/machine_files/events/'

    def get_platform(self):
        return 'virtual'

    def setUp(self):
        LogMixin.unit_test = False
        super().setUp()

    def _output(self, what, start, end, num):
        print("{}: Duration {:.5f}ms  Per second: {:2f}".format(
            what,
            (1000 * (end - start) / num),
            num / (end - start)
        ))

    def _handler(self, **kwargs):
        del kwargs

    def _handler2(self, num, **kwargs):
        del kwargs
        self.machine.events.post("test5_{}".format(num))

    def _benchmark(self, function, name, num=10000, iterations=10):
        function(num, -1)
        total = 0
        for i in range(iterations):
            start = time.time()
            function(num, i)
            end = time.time()
            total += (end - start) / num
            self._output(name, start, end, num)

        print("Total average {:.5f}ms".format(total * 1000 / iterations))
        return total/iterations

    def _unique_events(self, num, run):
        for i in range(num):
            self.machine.events.add_handler("test_{}".format(i + run * 10000), self._handler)

    def _multiple_events(self, num, run):
        for i in range(num):
            self.machine.events.add_handler("test_{}".format(i + run * 10000), self._handler)
            self.machine.events.add_handler("test_{}".format(i + run * 10000), self._handler)
            self.machine.events.add_handler("test_{}".format(i + run * 10000), self._handler)

    def _post_event(self, num, run):
        if run == -1:
            for i in range(num):
                self.machine.events.add_handler("test3_{}".format(i), self._handler)

        for i in range(num):
            self.machine.events.post("test3_{}".format(i))

        self.machine_run()

    def _post_event_with_new_handler(self, num, run):
        if run == -1:
            for i in range(num):
                self.machine.events.add_handler("test4_{}".format(i), self._handler2, num=i)
                self.machine.events.add_handler("test5_{}".format(i), self._handler)

        for i in range(num):
            self.machine.events.post("test4_{}".format(i), run=run)

        self.machine_run()

    def testEvents(self):
        num = 10000
        self._benchmark(self._unique_events, "Add unique event handlers", num, 10)
        self._benchmark(self._multiple_events, "Add multiple event handlers", int(num / 3), 10)
        self._benchmark(self._post_event, "Post events", num, 10)
        self._benchmark(self._post_event_with_new_handler, "Post events with new handler", num, 10)




from mpf.core.platform_controller import SwitchRuleSettings, DriverRuleSettings, PulseRuleSettings

from mpf.core.rgb_color import RGBColor
from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock, call
from mpf.platforms import p_roc_common


class MockPinProcModule(MagicMock):
    DriverCount = 256

    EventTypeAccelerometerIRQ = 11
    EventTypeAccelerometerX = 8
    EventTypeAccelerometerY = 9
    EventTypeAccelerometerZ = 10
    EventTypeBurstSwitchClosed = 7
    EventTypeBurstSwitchOpen = 6
    EventTypeDMDFrameDisplayed = 5
    EventTypeSwitchClosedDebounced = 1
    EventTypeSwitchClosedNondebounced = 3
    EventTypeSwitchOpenDebounced = 2
    EventTypeSwitchOpenNondebounced = 4

    MachineTypeCustom = 1
    MachineTypeInvalid = 0
    MachineTypePDB = 7
    MachineTypeSternSAM = 6
    MachineTypeSternWhitestar = 5
    MachineTypeWPC = 3
    MachineTypeWPC95 = 4
    MachineTypeWPCAlphanumeric = 2

    SwitchCount = 255
    SwitchNeverDebounceFirst = 192
    SwitchNeverDebounceLast = 255


class TestP3Roc(MpfTestCase):
    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/p3_roc/'

    def get_platform(self):
        return False

    def read_data(self, module, address):
        if module not in self._memory or address not in self._memory[module]:
            return 0
        return self._memory[module][address]

    def wait_for_platform(self):
        self._sync_count += 1
        num = self._sync_count
        result = self.machine.default_platform.run_proc_cmd_sync("_sync", num)
        assert result[0] == "sync"
        assert result[1] == num

    def _mock_loop(self):
        super()._mock_loop()
        self.loop._wait_for_external_executor = True

    def _driver_state_pulse(self, driver, milliseconds):
        driver["state"] = 1
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = False
        driver["outputDriveTime"] = milliseconds
        driver["patterOnTime"] = 0
        driver["patterOffTime"] = 0
        driver["patterEnable"] = False
        driver["futureEnable"] = False
        return driver

    def _driver_state_disable(self, driver):
        driver["state"] = 0
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = False
        driver["outputDriveTime"] = 0
        driver["patterOnTime"] = 0
        driver["patterOffTime"] = 0
        driver["patterEnable"] = False
        driver["futureEnable"] = False
        return driver

    def _driver_state_patter(self, driver, millisecondsOn, millisecondsOff, originalOnTime, now):
        driver["state"] = True
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = not now
        driver["outputDriveTime"] = originalOnTime
        driver["patterOnTime"] = millisecondsOn
        driver["patterOffTime"] = millisecondsOff
        driver["patterEnable"] = True
        driver["futureEnable"] = False
        return driver

    def _driver_pulsed_patter(self, driver, millisecondsOn, millisecondsOff, milliseconds_overall_patter_time, now):
        driver["state"] = True
        driver["timeslots"] = 0
        driver["waitForFirstTimeSlot"] = not now
        driver["outputDriveTime"] = milliseconds_overall_patter_time
        driver["patterOnTime"] = millisecondsOn
        driver["patterOffTime"] = millisecondsOff
        driver["patterEnable"] = True
        driver["futureEnable"] = False
        return driver

    def setUp(self):
        self._sync_count = 0
        self.expected_duration = 2
        p_roc_common.PINPROC_IMPORTED = True
        p_roc_common.pinproc = MockPinProcModule()
        self.pinproc = MagicMock(return_value=True)
        p_roc_common.pinproc.PinPROC = MagicMock(return_value=self.pinproc)
        p_roc_common.pinproc.normalize_machine_type = MagicMock(return_value=7)
        p_roc_common.pinproc.decode = None  # should not be called and therefore fail
        p_roc_common.pinproc.driver_state_pulse = self._driver_state_pulse

        p_roc_common.pinproc.driver_state_pulsed_patter = self._driver_pulsed_patter

        p_roc_common.pinproc.driver_state_disable = self._driver_state_disable

        p_roc_common.pinproc.driver_state_patter = self._driver_state_patter

        self.pinproc.switch_get_states = MagicMock(return_value=[0, 1] + [0] * 100)
        self.pinproc.read_data = self.read_data
        self.pinproc.write_data = MagicMock(return_value=True)
        self.pinproc.flush = MagicMock(return_value=True)
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.pinproc.driver_update_group_config = MagicMock(return_value=True)
        self.pinproc.driver_update_global_config = MagicMock(return_value=True)
        self.pinproc.driver_update_state = MagicMock(return_value=True)
        self.pinproc.driver_pulse = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)
        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_disable = MagicMock(return_value=True)
        self.pinproc.reset = MagicMock(return_value=True)
        self.pinproc.get_events = MagicMock(return_value=[])

        self._memory = {
            0x00: {         # manager
                0x00: 0,            # chip id
                0x01: 0x0002000E,   # version
                0x03: 0x01FF,       # dip switches (will be overwritten by GPIO config)
                0x04: 0x80          # GPIO state (GPIO 7 active)
            },
            0x02: {         # switch controller
                0x1000: 0xA3,       # SW-16 Address 0 Reg 0
                0x1001: 0x00,       # SW-16 Address 0 Reg 1
                0x1040: 0xA3,       # SW-16 Address 1 Reg 0
                0x1041: 0x13,       # SW-16 Address 1 Reg 1
                0x1080: 0xA4,       # SW-16 Address 2 Reg 0
                0x1081: 0x00,       # SW-16 Address 2 Reg 1
            }
        }

        super().setUp()

        p_roc_common.pinproc.normalize_machine_type.assert_called_once_with("pdb")

    def test_platform(self):
        self._test_write_data_init()
        self._test_accelerometer()
        self._test_servo_via_i2c()
        self._test_digital_outputs()
        self._test_pulse()
        self._test_pdb_matrix_light()
        self._test_enable_exception()
        self._test_allow_enable_disable()
        self._test_hw_rule_pulse()
        self._test_hw_rule_pulse_inverted_switch()
        self._test_hw_rule_pulse_disable_on_release()
        self._test_hw_rule_hold_pwm()
        self._test_hw_rule_hold_allow_enable()
        self._test_hw_rule_multiple_pulse()
        self._test_pd_led_servo()
        self._test_initial_switches()
        self._test_switches()
        self._test_flipper_single_coil()
        self._test_flipper_two_coils()
        self._test_flipper_two_coils_with_eos()
        self._test_flipper_one_coil_with_eos()
        self._test_pdb_gi_light()
        self._test_hw_rule_hold_no_allow_enable()
        self._test_leds()
        self._test_leds_inverted()
        self._test_steppers()
        self._test_driver_bank_config()

        # test hardware scan
        info_str = """Firmware Version: 2 Firmware Revision: 14 Hardware Board ID: 1
SW-16 boards found:
 - Board: 0 Switches: 16 Device Type: A3 Board ID: 0
 - Board: 1 Switches: 16 Device Type: A3 Board ID: 13
 - Board: 2 Switches: 16 Device Type: A4 Board ID: 0
"""
        self.assertEqual(info_str, self.machine.default_platform.get_info_string())
        self.assertEqual(2, self.machine.variables["p_roc_version"])
        self.assertEqual(14, self.machine.variables["p_roc_revision"])
        self.assertEqual(1, self.machine.variables["p_roc_hardware_version"])

    def _test_write_data_init(self):
        """All tests which assert init stuff."""
        # verify init of accelerometer
        self.pinproc.write_data.assert_has_calls([
            call(6, 0x0000, 0x000F),
            call(6, 0x012A, 0x0000),
            call(6, 0x010E, 0x0000),
            call(6, 0x012A, 0x0005),
            call(6, 0x012B, 0x0002),
            call(6, 0x0000, 0x1E0F)
        ])

        # assert on init of servo controller on i2c
        self.pinproc.write_data.assert_has_calls([
            call(7, 0x8000, 0x11),
            call(7, 0x8001, 0x04),
            call(7, 0x80FE, 130),
            call(7, 0x8000, 0x01)
        ], any_order=True)

        # assert on init of gpios
        self.pinproc.write_data.assert_has_calls([
            call(0, 3, 0x2600),
        ], any_order=True)

        # reset write_data
        self.pinproc.write_data.reset_mock()

    def _test_pulse(self):
        self.assertEqual("PD-16 Board 1 Bank 1", self.machine.coils["c_test"].hw_driver.get_board_name())
        # pulse coil A1-B1-2
        self.machine.coils["c_test"].pulse()
        self.wait_for_platform()
        number = self.machine.coils["c_test"].hw_driver.number
        self.pinproc.driver_pulse.assert_called_with(number, 23)
        assert not self.pinproc.driver_schedule.called

        self.machine.coils["c_sling_pulse_power"].pulse()
        self.wait_for_platform()
        number = self.machine.coils["c_sling_pulse_power"].hw_driver.number
        self.pinproc.driver_pulsed_patter.assert_called_with(number, 1, 1, 12, True)

    def _test_enable_exception(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.coils["c_test"].enable()

    def _test_allow_enable_disable(self):
        self.machine.coils["c_test_allow_enable"].enable()
        self.wait_for_platform()
        number = self.machine.coils["c_test_allow_enable"].hw_driver.number
        self.pinproc.driver_schedule.assert_called_with(
            number, 0xffffffff, 0, True)

        self.machine.coils["c_test_allow_enable"].disable()
        self.wait_for_platform()
        self.pinproc.driver_disable.assert_called_with(number)

    def _test_hw_rule_pulse(self):
        coil_number = self.machine.coils["c_slingshot_test"].hw_driver.number
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.wait_for_platform()
        self.machine.autofires["ac_slingshot_test"].enable()
        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(40, 'open_nondebounced', {'notifyHost': False, 'reloadActive': True}, [], False),
            call(40, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': True},
                 [{'futureEnable': False, 'patterOffTime': 0, 'polarity': True, 'waitForFirstTimeSlot': False,
                   'timeslots': 0, 'patterOnTime': 0, 'outputDriveTime': 10, 'patterEnable': False, 'state': 1,
                   'driverNum': coil_number}], False),
            call(40, 'open_debounced', {'notifyHost': True, 'reloadActive': True}, [], False),
            call(40, 'closed_debounced', {'notifyHost': True, 'reloadActive': True}, [], False),
        ], any_order=True)

        self.pinproc.switch_update_rule = MagicMock(return_value=True)

        # test disable
        self.machine.autofires["ac_slingshot_test"].disable()
        self.wait_for_platform()

        self.pinproc.switch_update_rule.assert_has_calls([
            call(40, 'open_nondebounced', {'notifyHost': False, 'reloadActive': True}, []),
            call(40, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': True}, []),
            call(40, 'open_debounced', {'notifyHost': True, 'reloadActive': True}, []),
            call(40, 'closed_debounced', {'notifyHost': True, 'reloadActive': True}, []),
        ], any_order=True)

        self.pinproc.driver_disable.assert_called_with(coil_number)

        # sling with pulse power
        coil_number = self.machine.coils["c_sling_pulse_power"].hw_driver.number
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.wait_for_platform()
        self.machine.autofires["ac_sling_pulse_power"].enable()
        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(66, 'open_nondebounced', {'notifyHost': False, 'reloadActive': True}, [], False),
            call(66, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': True},
                 [{'driverNum': coil_number, 'outputDriveTime': 12, 'polarity': True, 'state': True,
                   'waitForFirstTimeSlot': False, 'timeslots': 0, 'patterOnTime': 1, 'patterOffTime': 1,
                   'patterEnable': True, 'futureEnable': False}], False),
            call(66, 'open_debounced', {'notifyHost': True, 'reloadActive': True}, [], False),
            call(66, 'closed_debounced', {'notifyHost': True, 'reloadActive': True}, [], False),
        ], any_order=True)

    def _test_hw_rule_pulse_inverted_switch(self):
        coil_number = self.machine.coils["c_coil_pwm_test"].hw_driver.number
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.wait_for_platform()
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.machine.autofires["ac_switch_nc_test"].enable()
        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(41, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': True}, [], False),
            call(41, 'open_nondebounced', {'notifyHost': False, 'reloadActive': True},
                 [{'futureEnable': False, 'patterOffTime': 0, 'polarity': True, 'waitForFirstTimeSlot': False,
                   'timeslots': 0, 'patterOnTime': 0, 'outputDriveTime': 10, 'patterEnable': False, 'state': 1,
                   'driverNum': coil_number}], False),
            call(41, 'open_debounced', {'notifyHost': True, 'reloadActive': True}, [], False),
            call(41, 'closed_debounced', {'notifyHost': True, 'reloadActive': True}, [], False),
        ], any_order=True)

        # test disable
        self.machine.autofires["ac_switch_nc_test"].disable()
        self.wait_for_platform()
        self.pinproc.driver_disable.assert_called_with(coil_number)

    def _test_hw_rule_pulse_disable_on_release(self):
        coil_number = self.machine.coils["c_test"].hw_driver.number
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.wait_for_platform()
        rule = self.machine.platform_controller.set_pulse_on_hit_and_release_rule(
            SwitchRuleSettings(switch=self.machine.switches["s_test"], debounce=True, invert=False),
            DriverRuleSettings(driver=self.machine.coils["c_test"], recycle=False),
            PulseRuleSettings(duration=23, power=1.0))
        self.wait_for_platform()

        self.pinproc.switch_update_rule.assert_has_calls([
            call(23, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(23, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(23, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, [
                {'patterOnTime': 0, 'driverNum': coil_number, 'waitForFirstTimeSlot': False, 'patterOffTime': 0,
                 'polarity': True,
                 'timeslots': 0, 'state': 0, 'patterEnable': False, 'outputDriveTime': 0, 'futureEnable': False}],
                 False),
            call(23, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, [
                {'patterOnTime': 0, 'driverNum': coil_number, 'waitForFirstTimeSlot': False, 'patterOffTime': 0,
                 'polarity': True,
                 'timeslots': 0, 'state': 1, 'patterEnable': False, 'outputDriveTime': 23, 'futureEnable': False}],
                 False),
        ], any_order=True)

        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.machine.platform_controller.clear_hw_rule(rule)
        self.wait_for_platform()

        self.pinproc.switch_update_rule.assert_has_calls([
            call(23, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, []),
            call(23, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, []),
            call(23, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, []),
            call(23, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, []),
        ], any_order=True)

    def _test_hw_rule_hold_pwm(self):
        return  # currently not cupported
        self.machine.default_platform.set_hw_rule(
                sw_name="s_test",
                sw_activity=1,
                driver_name="c_coil_pwm_test",
                driver_action='hold',
                disable_on_release=False)

        self.pinproc.switch_update_rule.assert_has_calls([
            call(
                23, 'closed_debounced',
                {'notifyHost': True, 'reloadActive': False},
                [{'patterEnable': False,
                  'patterOnTime': 0,
                  'timeslots': 0,
                  'futureEnable': False,
                  'state': False,
                  'patterOffTime': 0,
                  'outputDriveTime': 0,
                  'driverNum': 11,
                  'polarity': True,
                  'waitForFirstTimeSlot': False}],
                False),
        ])

        p_roc_common.pinproc.driver_state_patter.assert_called_with(8, 2, 8, 0, True)

        # now add disable rule
        self.machine.default_platform.set_hw_rule(
                sw_name="s_test",
                sw_activity=0,
                driver_name="c_coil_pwm_test",
                driver_action='disable',
                disable_on_release=False)

        self.pinproc.switch_update_rule.assert_has_calls([
            call(
                23, 'open_debounced',
                {'notifyHost': True, 'reloadActive': False},
                [{'patterEnable': False,
                  'patterOnTime': 0,
                  'timeslots': 0,
                  'futureEnable': False,
                  'state': False,
                  'patterOffTime': 0,
                  'outputDriveTime': 0,
                  'driverNum': 10,
                  'polarity': True,
                  'waitForFirstTimeSlot': False}],
                False),
        ], any_order=True)

        p_roc_common.pinproc.driver_state_disable.assert_called_with(8)

        self.machine.default_platform.clear_hw_rule("s_test", "c_coil_pwm_test")

    def _test_hw_rule_hold_allow_enable(self):
        coil_number = self.machine.coils["c_test_allow_enable"].hw_driver.number
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.wait_for_platform()
        rule = self.machine.platform_controller.set_pulse_on_hit_and_enable_and_release_rule(
            SwitchRuleSettings(switch=self.machine.switches["s_test"], debounce=True, invert=False),
            DriverRuleSettings(driver=self.machine.coils["c_test_allow_enable"], recycle=False),
            PulseRuleSettings(duration=23, power=1.0))

        self.wait_for_platform()

        self.pinproc.switch_update_rule.assert_has_calls([
            call(23, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(23, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(23, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, [
                {'patterOnTime': 0, 'driverNum': coil_number, 'waitForFirstTimeSlot': False, 'patterOffTime': 0,
                 'polarity': True,
                 'timeslots': 0, 'state': 0, 'patterEnable': False, 'outputDriveTime': 0, 'futureEnable': False}],
                 False),
            call(23, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, [
                {'patterOnTime': 0, 'driverNum': coil_number, 'waitForFirstTimeSlot': False, 'patterOffTime': 0,
                 'polarity': True,
                 'timeslots': 0, 'state': 1, 'patterEnable': False, 'outputDriveTime': 0, 'futureEnable': False}],
                 False),
        ], any_order=True)

        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.machine.platform_controller.clear_hw_rule(rule)
        self.wait_for_platform()

        self.pinproc.switch_update_rule.assert_has_calls([
            call(23, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, []),
            call(23, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, []),
            call(23, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, []),
            call(23, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, []),
        ], any_order=True)


    def _test_hw_rule_hold_no_allow_enable(self):
        # enable coil which does not have allow_enable
        with self.assertRaises(AssertionError):
            self.machine.platform_controller.set_pulse_on_hit_and_enable_and_release_rule(
                SwitchRuleSettings(switch=self.machine.switches["s_test"], debounce=True, invert=False),
                DriverRuleSettings(driver=self.machine.coils["c_test"], recycle=False),
                PulseRuleSettings(duration=23, power=1.0))

    def _test_hw_rule_multiple_pulse(self):
        coil_number = self.machine.coils["c_test"].hw_driver.number
        coil_number2 = self.machine.coils["c_coil_pwm_test"].hw_driver.number
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.wait_for_platform()
        self.machine.platform_controller.set_pulse_on_hit_rule(
            SwitchRuleSettings(switch=self.machine.switches["s_test"], debounce=True, invert=False),
            DriverRuleSettings(driver=self.machine.coils["c_test"], recycle=False),
            PulseRuleSettings(duration=23, power=1.0))

        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(23, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(23, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(23, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(23, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, [
                {'patterOnTime': 0, 'driverNum': coil_number, 'waitForFirstTimeSlot': False, 'patterOffTime': 0,
                 'polarity': True,
                 'timeslots': 0, 'state': 1, 'patterEnable': False, 'outputDriveTime': 23, 'futureEnable': False}],
                 False),
        ], any_order=True)

        self.pinproc.switch_update_rule = MagicMock(return_value=True)

        # test setting the same rule again
        self.machine.platform_controller.set_pulse_on_hit_rule(
            SwitchRuleSettings(switch=self.machine.switches["s_test"], debounce=True, invert=False),
            DriverRuleSettings(driver=self.machine.coils["c_test"], recycle=False),
            PulseRuleSettings(duration=23, power=1.0))
        self.wait_for_platform()

        self.pinproc.switch_update_rule.assert_has_calls([
            call(23, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(23, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(23, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(23, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, [
                {'patterOnTime': 0, 'driverNum': coil_number, 'waitForFirstTimeSlot': False, 'patterOffTime': 0,
                 'polarity': True,
                 'timeslots': 0, 'state': 1, 'patterEnable': False, 'outputDriveTime': 23, 'futureEnable': False}],
                 False),
        ], any_order=True)

        self.machine.platform_controller.set_pulse_on_hit_rule(
            SwitchRuleSettings(switch=self.machine.switches["s_test"], debounce=True, invert=False),
            DriverRuleSettings(driver=self.machine.coils["c_coil_pwm_test"], recycle=False),
            PulseRuleSettings(duration=23, power=1.0))

        self.wait_for_platform()

        self.pinproc.switch_update_rule.assert_has_calls([
            call(23, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(23, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(23, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(23, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, [
                {'patterOnTime': 0, 'driverNum': coil_number, 'waitForFirstTimeSlot': False, 'patterOffTime': 0,
                 'polarity': True,
                 'timeslots': 0, 'state': 1, 'patterEnable': False, 'outputDriveTime': 23, 'futureEnable': False},
                {'patterOnTime': 0, 'driverNum': coil_number2, 'waitForFirstTimeSlot': False, 'patterOffTime': 0,
                 'polarity': True,
                 'timeslots': 0, 'state': 1, 'patterEnable': False, 'outputDriveTime': 23, 'futureEnable': False},
            ],
                 False),
        ], any_order=True)

    def _test_servo_via_i2c(self):
        self.pinproc.write_data = MagicMock(return_value=True)
        self.machine.servos["servo1"].go_to_position(0)
        self.wait_for_platform()

        self.pinproc.write_data.assert_has_calls([
            call(7, 0x8012, 0),
            call(7, 0x8013, 0),
            call(7, 0x8014, 150),
            call(7, 0x8015, 0)
        ])
        self.pinproc.write_data = MagicMock(return_value=True)
        self.machine.servos["servo1"].go_to_position(1)
        self.wait_for_platform()

        self.pinproc.write_data.assert_has_calls([
            call(7, 0x8012, 0),
            call(7, 0x8013, 0),
            call(7, 0x8014, 88),
            call(7, 0x8015, 2)
        ])

    def _test_pd_led_servo(self):
        self.pinproc.write_data = MagicMock(return_value=True)
        self.machine.servos["servo_pd_led_0"].go_to_position(0)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 72),              # low byte of address (72)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 127),  # set servo position
            ], False)

        self.pinproc.write_data = MagicMock(return_value=True)
        self.machine.servos["servo_pd_led_0"].go_to_position(1)
        self.wait_for_platform()

        self.pinproc.write_data.assert_has_calls([
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 72),              # low byte of address (72)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 255),  # set servo position
            ], False)

    def _test_initial_switches(self):
        self.assertSwitchState("s_test", 0)
        self.assertSwitchState("s_test_000", 0)
        self.assertSwitchState("s_flipper", 1)
        self.assertSwitchState("s_gpio0", 0)
        self.assertSwitchState("s_gpio7", 1)

    def _test_switches(self):
        self.assertSwitchState("s_test", 0)
        # closed debounced -> switch active
        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 1, 'value': 23}])
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.assertSwitchState("s_test", 1)

        # open debounces -> inactive
        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 2, 'value': 23}])
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.assertSwitchState("s_test", 0)

        self.assertSwitchState("s_test_no_debounce", 0)
        # closed non debounced -> should be active
        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 3, 'value': 24}])
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.assertSwitchState("s_test_no_debounce", 1)

        # open non debounced -> should be inactive
        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 4, 'value': 24}])
        self.wait_for_platform()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.assertSwitchState("s_test_no_debounce", 0)

        self.pinproc.get_events = MagicMock(return_value=[])

        # test gpio changes
        self._memory[0][4] = 0x01   # gpio 7 low, gpio 0 high
        self.advance_time_and_run(.01)
        self.wait_for_platform()
        self.advance_time_and_run(.01)
        self.assertSwitchState("s_gpio0", 1)
        self.assertSwitchState("s_gpio7", 0)

    def _test_accelerometer(self):
        self.machine.accelerometers["p3_roc_accelerometer"].update_acceleration = MagicMock(return_value=True)

        # process accelerometer event
        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 8, 'value': 4096},
            {'type': 9, 'value': 0},
            {'type': 10, 'value': 8192}
        ])
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.advance_time_and_run(.1)

        # check correct decoding of 2 complement
        self.machine.accelerometers["p3_roc_accelerometer"].update_acceleration.assert_called_with(1.0, 0.0, -2.0)

        self.pinproc.get_events = MagicMock(return_value=[])

    def _test_flipper_single_coil(self):
        coil_number = self.machine.coils["c_flipper_main"].hw_driver.number
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.wait_for_platform()
        # enable
        self.machine.flippers["f_test_single"].enable()

        self.wait_for_platform()

        self.pinproc.switch_update_rule.assert_has_calls([
            call(1, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': False}, [
                {'timeslots': 0, 'waitForFirstTimeSlot': False, 'polarity': True, 'driverNum': coil_number,
                 'outputDriveTime': 11,
                 'futureEnable': False, 'patterEnable': True, 'patterOnTime': 3, 'patterOffTime': 5, 'state': True}],
                 False),
            call(1, 'open_debounced', {'notifyHost': True, 'reloadActive': False}, [], False),
            call(1, 'closed_debounced', {'notifyHost': True, 'reloadActive': False}, [], False),
            call(1, 'open_nondebounced', {'notifyHost': False, 'reloadActive': False}, [
                {'timeslots': 0, 'waitForFirstTimeSlot': False, 'polarity': True, 'driverNum': coil_number,
                 'outputDriveTime': 0,
                 'futureEnable': False, 'patterEnable': False, 'patterOnTime': 0, 'patterOffTime': 0, 'state': 0}],
                 False),
        ], any_order=True)

        # disable
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.machine.flippers["f_test_single"].disable()
        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(1, 'open_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(1, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(1, 'open_debounced', {'notifyHost': True, 'reloadActive': False}, []),
            call(1, 'closed_debounced', {'notifyHost': True, 'reloadActive': False}, []),
        ], any_order=True)
        self.assertEqual(4, self.pinproc.switch_update_rule.call_count)

    def _test_flipper_two_coils(self):
        coil_number = self.machine.coils["c_flipper_main"].hw_driver.number
        coil_number2 = self.machine.coils["c_flipper_hold"].hw_driver.number
        # we pulse the main coil (20)
        # hold coil (21) is pulsed + enabled
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.machine.flippers["f_test_hold"].enable()
        self.wait_for_platform()

        self.pinproc.switch_update_rule.assert_has_calls([
            call(1, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, [
                {'patterEnable': False, 'waitForFirstTimeSlot': False, 'state': 0, 'timeslots': 0, 'patterOffTime': 0,
                 'outputDriveTime': 0, 'driverNum': coil_number, 'polarity': True, 'patterOnTime': 0,
                 'futureEnable': False},
                {'patterEnable': False, 'waitForFirstTimeSlot': False, 'state': 0, 'timeslots': 0, 'patterOffTime': 0,
                 'outputDriveTime': 0, 'driverNum': coil_number2, 'polarity': True, 'patterOnTime': 0,
                 'futureEnable': False}],
                 False),
            call(1, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(1, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, [
                {'patterEnable': False, 'waitForFirstTimeSlot': False, 'state': 1, 'timeslots': 0, 'patterOffTime': 0,
                 'outputDriveTime': 10, 'driverNum': coil_number, 'polarity': True, 'patterOnTime': 0,
                 'futureEnable': False},
                {'patterEnable': True, 'waitForFirstTimeSlot': False, 'state': True, 'timeslots': 0, 'patterOffTime': 7,
                 'outputDriveTime': 10, 'driverNum': coil_number2, 'polarity': True, 'patterOnTime': 1,
                 'futureEnable': False}],
                 False),
            call(1, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
        ], any_order=True)

        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.machine.flippers["f_test_hold"].disable()
        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(1, 'open_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(1, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(1, 'open_debounced', {'notifyHost': True, 'reloadActive': False}, []),
            call(1, 'closed_debounced', {'notifyHost': True, 'reloadActive': False}, []),
        ], any_order=True)

    def _test_flipper_two_coils_with_eos(self):
        coil_number = self.machine.coils["c_flipper_main"].hw_driver.number
        coil_number2 = self.machine.coils["c_flipper_hold"].hw_driver.number
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.machine.flippers["f_test_hold_eos"].enable()
        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(2, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(2, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(2, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(2, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, [
                {'patterOnTime': 0, 'outputDriveTime': 0, 'timeslots': 0, 'patterOffTime': 0, 'polarity': True,
                 'driverNum': coil_number, 'state': 0, 'futureEnable': False, 'patterEnable': False,
                 'waitForFirstTimeSlot': False}], False),
            call(1, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, [
                {'patterOnTime': 0, 'outputDriveTime': 0, 'timeslots': 0, 'patterOffTime': 0, 'polarity': True,
                 'driverNum': coil_number, 'state': 0, 'futureEnable': False, 'patterEnable': False,
                 'waitForFirstTimeSlot': False},
                {'patterOnTime': 0, 'outputDriveTime': 0, 'timeslots': 0, 'patterOffTime': 0, 'polarity': True,
                 'driverNum': coil_number2, 'state': 0, 'futureEnable': False, 'patterEnable': False,
                 'waitForFirstTimeSlot': False}], False),
            call(1, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(1, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(1, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, [
                {'patterOnTime': 0, 'outputDriveTime': 10, 'timeslots': 0, 'patterOffTime': 0, 'polarity': True,
                 'driverNum': coil_number, 'state': True, 'futureEnable': False, 'patterEnable': False,
                 'waitForFirstTimeSlot': False},
                {'patterOnTime': 1, 'outputDriveTime': 10, 'timeslots': 0, 'patterOffTime': 7, 'polarity': True,
                 'driverNum': coil_number2, 'state': True, 'futureEnable': False, 'patterEnable': True,
                 'waitForFirstTimeSlot': False}], False)
        ], any_order=True)

        # disable
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.machine.flippers["f_test_hold_eos"].disable()
        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(1, 'open_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(1, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(1, 'open_debounced', {'notifyHost': True, 'reloadActive': False}, []),
            call(1, 'closed_debounced', {'notifyHost': True, 'reloadActive': False}, []),
            call(2, 'open_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(2, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(2, 'open_debounced', {'notifyHost': True, 'reloadActive': False}, []),
            call(2, 'closed_debounced', {'notifyHost': True, 'reloadActive': False}, []),
        ], any_order=True)

    def _test_flipper_one_coil_with_eos(self):
        # single coil with eos
        coil_number = self.machine.coils["c_flipper_main"].hw_driver.number
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.machine.flippers["f_test_single_eos"].enable()
        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(2, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, [], False),
            call(2, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(2, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(2, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': False}, [
                {'driverNum': coil_number, 'outputDriveTime': 0, 'polarity': True, 'state': True,
                 'waitForFirstTimeSlot': False, 'timeslots': 0, 'patterOnTime': 3, 'patterOffTime': 5,
                 'patterEnable': True, 'futureEnable': False}], False),
            call(1, 'open_nondebounced', {'reloadActive': False, 'notifyHost': False}, [
                {'patterOnTime': 0, 'outputDriveTime': 0, 'timeslots': 0, 'patterOffTime': 0, 'polarity': True,
                 'driverNum': coil_number, 'state': 0, 'futureEnable': False, 'patterEnable': False,
                 'waitForFirstTimeSlot': False}], False),
            call(1, 'closed_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(1, 'open_debounced', {'reloadActive': False, 'notifyHost': True}, [], False),
            call(1, 'closed_nondebounced', {'reloadActive': False, 'notifyHost': False}, [
                {'patterOnTime': 3, 'outputDriveTime': 10, 'timeslots': 0, 'patterOffTime': 5, 'polarity': True,
                 'driverNum': coil_number, 'state': True, 'futureEnable': False, 'patterEnable': True,
                 'waitForFirstTimeSlot': False}], False),
        ], any_order=True)

        # disable
        self.pinproc.switch_update_rule = MagicMock(return_value=True)
        self.machine.flippers["f_test_single_eos"].disable()
        self.wait_for_platform()
        self.pinproc.switch_update_rule.assert_has_calls([
            call(1, 'open_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(1, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(1, 'open_debounced', {'notifyHost': True, 'reloadActive': False}, []),
            call(1, 'closed_debounced', {'notifyHost': True, 'reloadActive': False}, []),
            call(2, 'open_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(2, 'closed_nondebounced', {'notifyHost': False, 'reloadActive': False}, []),
            call(2, 'open_debounced', {'notifyHost': True, 'reloadActive': False}, []),
            call(2, 'closed_debounced', {'notifyHost': True, 'reloadActive': False}, []),
        ], any_order=True)

    def _test_pdb_matrix_light(self):
        # very simple check for matrix config
        self.pinproc.driver_update_group_config.assert_has_calls(
            [call(4, 100, 5, 0, 0, True, True, True, True)]
        )

        # test enable of matrix light
        assert not self.pinproc.driver_patter.called
        assert not self.pinproc.driver_schedule.called
        self.machine.lights["test_pdb_light"].on()
        self.advance_time_and_run(.02)
        self.wait_for_platform()
        self.pinproc.driver_schedule.assert_called_with(
            32, 4294967295, 0, True
        )

        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.machine.lights["test_pdb_light"].on(brightness=128)
        self.advance_time_and_run(.02)
        self.wait_for_platform()
        self.pinproc.driver_patter.assert_called_with(
            32, 1, 1, 0, True
        )

        # test disable of matrix light
        assert not self.pinproc.driver_disable.called
        self.machine.lights["test_pdb_light"].off()
        self.advance_time_and_run(.02)
        self.wait_for_platform()
        self.pinproc.driver_disable.assert_called_with(32)

    def _test_pdb_gi_light(self):
        # test gi on
        device = self.machine.lights["test_gi"]
        num = self.machine.coils["test_gi"].hw_driver.number
        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)
        device.color("white")
        self.wait_for_platform()
        self.pinproc.driver_schedule.assert_has_calls([
            call(num, 4294967295, 0, True)])
        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)

        device.color([128, 128, 128])
        self.wait_for_platform()
        self.pinproc.driver_patter.assert_has_calls([
            call(num, 1, 1, 0, True)])
        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)

        device.color([245, 245, 245])
        self.wait_for_platform()
        self.pinproc.driver_patter.assert_has_calls([
            call(num, 19, 1, 0, True)])
        self.pinproc.driver_patter = MagicMock(return_value=True)
        self.pinproc.driver_schedule = MagicMock(return_value=True)

        # test gi off
        self.pinproc.driver_disable = MagicMock(return_value=True)
        device.color("off")
        self.wait_for_platform()
        self.pinproc.driver_disable.assert_has_calls([
            call(num)])

    def _test_leds(self):
        device = self.machine.lights["test_led"]
        self.pinproc.write_data = MagicMock(return_value=True)

        # test led on
        device.on()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            # first LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 1),               # low byte of address (1)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 255),  # set color (255)
            # second LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 255),  # set color (255)
            # third LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 255),  # set color (255)
            ], False)
        self.pinproc.write_data = MagicMock(return_value=True)

        # test led off
        device.off()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            # first LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 1),               # low byte of address (1)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 0),    # set color (0)
            # second LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 0),    # set color (0)
            # third LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 0),    # set color (0)
            ], False)
        self.pinproc.write_data = MagicMock(return_value=True)

        # test led color
        device.color(RGBColor((2, 23, 42)))
        self.advance_time_and_run(.1)
        self.wait_for_platform()

        self.pinproc.write_data.assert_has_calls([
            # first LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 1),               # low byte of address (1)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 2),    # set color (2)
            # second LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 23),   # set color (23)
            # third LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 42),   # set color (42)
            ], False)
        self.pinproc.write_data = MagicMock(return_value=True)

        device = self.machine.lights["test_led2"]
        device.on()
        self.advance_time_and_run(1)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            # first LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 7),               # low byte of address (1)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 255),  # set color (255)
            # second LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 255),  # set color (255)
            # third LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 255),  # set color (255)
            ], False)
        self.pinproc.write_data = MagicMock(return_value=True)

        # test led color with fade
        device.color(RGBColor((13, 37, 238)), fade_ms=42)
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            # fade ms
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (3 << 8) | 10),   # set fade lower (42/4 = 10)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (4 << 8) | 0),    # set fade higher (0)
            # first LED addr
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 7),               # low byte of address (7)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (2 << 8) | 13),   # set color (13)
            # second LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (2 << 8) | 37),   # set color (37)
            # third LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (2 << 8) | 238),  # set color (238)
            ], False)

        self.advance_time_and_run(.5)
        self.pinproc.write_data = MagicMock(return_value=True)

        # test led color with fade
        device.color(RGBColor((255, 37, 238)), fade_ms=20)
        self.machine.lights["test_led3"].color(RGBColor((253, 0, 0)), fade_ms=20)
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            # fade ms
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (3 << 8) | 5),    # set fade lower (20/4 = 5)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (4 << 8) | 0),    # set fade higher (0)
            # first LED addr
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 7),               # low byte of address (7)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (2 << 8) | 255),   # set color (13)
            # second LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (2 << 8) | 37),   # set color (37)
            # third LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (2 << 8) | 238),  # set color (238)
            # forth LED (test_led3)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (2 << 8) | 253),  # set color (253)
            # fifth LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (2 << 8) | 0),  # set color (0)
            # sixth LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (2 << 8) | 0),  # set color (0)
            ], False)

    def _test_leds_inverted(self):
        device = self.machine.lights["test_led_inverted"]
        self.pinproc.write_data = MagicMock(return_value=True)
        # test led on
        device.on()
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            # first LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 4),               # low byte of address (4)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 0),    # set color (0)
            # second LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 0),    # set color (0)
            # third LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 0),    # set color (0)
            ], False)
        self.pinproc.write_data = MagicMock(return_value=True)

        # test led off
        device.color("off")
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            # first LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 4),               # low byte of address (4)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 255),  # set color (255)
            # second LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 255),  # set color (255)
            # third LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 255),  # set color (255)
            ], False)
        self.pinproc.write_data = MagicMock(return_value=True)

        # test led color
        device.color(RGBColor((2, 23, 42)))
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            # first LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | 4),               # low byte of address (4)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (6 << 8)),        # high byte of address (0)
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 253),  # set color (255 - 2)
            # second LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 232),  # set color (255 - 23)
            # third LED
            call(3, 3072, 0x01000000 | (2 & 0x3F) << 16 | (1 << 8) | 213),  # set color (255 - 42)
            ], False)
        self.pinproc.write_data = MagicMock(return_value=True)

    def _test_steppers(self):
        stepper1 = self.machine.steppers["stepper1"]
        stepper2 = self.machine.steppers["stepper2"]

        self.pinproc.get_events = MagicMock(return_value=[
            {'type': 1, 'value': 64}, {'type': 1, 'value': 65}])
        self.wait_for_platform()
        self.advance_time_and_run(.01)
        self.wait_for_platform()
        self.assertSwitchState("s_stepper1_home", 1)
        self.assertSwitchState("s_stepper2_home", 1)
        self.pinproc.get_events = MagicMock(return_value=[])

        # test stepper 1
        self.pinproc.write_data = MagicMock()
        stepper1._move_to_absolute_position(11)
        self.advance_time_and_run(.1)
        self.wait_for_platform()

        self.pinproc.write_data.assert_has_calls([
            call(3, 3072, 0x1040000 + 11),
            call(3, 3072, 0x1040600),
            call(3, 3072, 0x1040700 + 23)
        ], True)

        # test stepper 2
        self.pinproc.write_data = MagicMock()
        stepper2._move_to_absolute_position(500)
        self.advance_time_and_run(.1)
        self.wait_for_platform()

        self.pinproc.write_data.assert_has_calls([
            call(3, 3072, 0x1040000 + (500 & 0xFF)),
            call(3, 3072, 0x1040600 + ((500 >> 8) & 0xFF)),
            call(3, 3072, 0x1040700 + 24)
        ], True)

        # move again. it should wait
        self.pinproc.write_data = MagicMock()
        stepper2._move_to_absolute_position(450)
        self.advance_time_and_run(.1)
        self.wait_for_platform()
        self.assertEqual(0, self.pinproc.write_data.call_count)

        self.advance_time_and_run(1)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            call(3, 3072, 0x1040000 + 50),
            call(3, 3072, 0x1040600 + (1 << 7)),
            call(3, 3072, 0x1040700 + 24)
        ], True)

    def _test_digital_outputs(self):
        self.pinproc.write_data.reset_mock()

        self.machine.digital_outputs["d_gpio1"].enable()
        self.machine.digital_outputs["d_gpio5"].enable()
        self.advance_time_and_run(.01)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            call(0, 3, 0x2602),
            call(0, 3, 0x2622),
        ], any_order=False)

        self.pinproc.write_data.reset_mock()

        self.machine.digital_outputs["d_gpio1"].disable()
        self.advance_time_and_run(.01)
        self.wait_for_platform()
        self.pinproc.write_data.assert_has_calls([
            call(0, 3, 0x2620),
        ], any_order=False)

    def _test_driver_bank_config(self):
        """Check configured banks."""
        configured_banks = set()
        enabled_banks = []
        for call in self.pinproc.driver_update_group_config.mock_calls[4:]:
            configured_banks.add(call[1][2])
            if call[1][7]:
                enabled_banks.append(call[1][2])

        self.assertEqual({0, 1, 2, 3, 4, 5, 10, 11, 12, 13}, configured_banks)
        self.assertEqual([0, 1, 2, 3, 4, 5, 10, 11, 12, 13], sorted(enabled_banks))



"""Support for physical RGB DMDs."""
from functools import partial

from mpf.core.machine import MachineController
from mpf.core.platform import RgbDmdPlatform

from mpf.core.system_wide_device import SystemWideDevice


class RgbDmd(SystemWideDevice):

    """A physical DMD."""

    config_section = 'rgb_dmds'
    collection = 'rgb_dmds'
    class_label = 'rgb_dmd'

    __slots__ = ["hw_device", "platform"]

    @classmethod
    def device_class_init(cls, machine: MachineController):
        """Create BCP methods.

        Args:
        ----
            machine: MachineController which is used
        """
        machine.bcp.interface.register_command_callback("rgb_dmd_frame", partial(cls._bcp_receive_dmd_frame, machine))

    def __init__(self, machine, name):
        """Initialise DMD."""
        self.hw_device = None
        self.platform = None        # type: RgbDmdPlatform
        super().__init__(machine, name)

    async def _initialize(self):
        await super()._initialize()
        self.platform = self.machine.get_platform_sections("rgb_dmd", self.config['platform'])
        self.platform.assert_has_feature("rgb_dmds")
        self.hw_device = self.platform.configure_rgb_dmd(self.name)
        self._update_brightness(None)

    def _update_brightness(self, future):
        del future
        brightness, brightness_changed_future = self.config['hardware_brightness'].evaluate_and_subscribe([])
        self.hw_device.set_brightness(brightness)
        brightness_changed_future.add_done_callback(self._update_brightness)

    @classmethod
    async def _bcp_receive_dmd_frame(cls, machine, client, name, rawbytes, **kwargs):
        """Update dmd from BCP."""
        del client
        del kwargs

        if name not in machine.rgb_dmds:
            raise TypeError("rgb dmd {} not known".format(name))

        machine.rgb_dmds[name].update(rawbytes)

    def update(self, data: bytes):
        """Update data on the dmd.

        Args:
        ----
            data: bytes to send
        """
        self.hw_device.update(data)



"""Test event player."""
from collections import namedtuple

from mpf.tests.MpfTestCase import MpfTestCase


class TestEventPlayer(MpfTestCase):

    def get_config_file(self):
        return 'test_event_player.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/event_players/'

    def test_load_and_play(self):
        self.mock_event("event1")
        self.mock_event("event2")
        self.mock_event("event3")
        self.post_event("play_express_single")
        self.assertEqual(1, self._events['event1'])
        self.assertEqual(0, self._events['event2'])
        self.assertEqual(0, self._events['event3'])

        self.mock_event("event1")
        self.mock_event("event2")
        self.mock_event("event3")
        self.post_event("play_express_multiple")
        self.assertEqual(1, self._events['event1'])
        self.assertEqual(1, self._events['event2'])
        self.assertEqual(0, self._events['event3'])

        self.mock_event("event1")
        self.mock_event("event2")
        self.mock_event("event3")
        self.post_event("play_single_list")
        self.assertEqual(1, self._events['event1'])
        self.assertEqual(0, self._events['event2'])
        self.assertEqual(0, self._events['event3'])

        self.mock_event("event1")
        self.mock_event("event2")
        self.mock_event("event3")
        self.post_event("play_single_string")
        self.assertEqual(1, self._events['event1'])
        self.assertEqual(0, self._events['event2'])
        self.assertEqual(0, self._events['event3'])

        self.mock_event("event1")
        self.mock_event("event2")
        self.mock_event("event3")
        self.post_event("play_multiple_list")
        self.assertEqual(1, self._events['event1'])
        self.assertEqual(1, self._events['event2'])
        self.assertEqual(1, self._events['event3'])

        self.mock_event("event1")
        self.mock_event("event2")
        self.mock_event("event3")
        self.post_event("play_multiple_string")
        self.assertEqual(1, self._events['event1'])
        self.assertEqual(1, self._events['event2'])
        self.assertEqual(1, self._events['event3'])

        self.mock_event("event1")
        self.mock_event("event2")
        self.mock_event("event3")
        self.post_event("play_multiple_args")
        self.assertEqual(1, self._events['event1'])
        self.assertEqual({"a": "b", "priority": 0}, self._last_event_kwargs['event1'])
        self.assertEqual(1, self._events['event2'])
        self.assertEqual({"priority": 0}, self._last_event_kwargs['event2'])
        self.assertEqual(1, self._events['event3'])
        self.assertEqual({"a": 1, "b": 2, "priority": 0}, self._last_event_kwargs['event3'])

        self.mock_event("event1")
        self.post_event("play_multiple_args2")
        self.assertEqual({"a": "b", "c": "d", "priority": 0}, self._last_event_kwargs['event1'])

        self.mock_event("event1")
        self.mock_event("event2")
        self.mock_event("event3")
        self.machine.shows['test_event_show'].play(loops=0)
        self.advance_time_and_run()
        self.assertEqual(1, self._events['event1'])
        self.assertEqual(1, self._events['event2'])
        self.assertEqual(1, self._events['event3'])

    def test_condition_and_priority(self):
        self.mock_event("condition_ok")
        self.mock_event("condition_ok2")
        self.mock_event("priority_ok")
        self.post_event("test_conditional")
        self.assertEventNotCalled("condition_ok")
        self.assertEventNotCalled("condition_ok2")
        self.assertEventCalled("priority_ok")

        arg_obj = namedtuple("Arg", ["abc"])
        arg = arg_obj(1)

        self.post_event_with_params("test_conditional", arg=arg)
        self.assertEventCalled("condition_ok")
        self.assertEventCalled("condition_ok2")

    def test_handler_condition(self):
        # test neither condition passing
        self.mock_event("event_always")
        self.mock_event("event_if_modeactive")
        self.mock_event("event_if_modestopping")

        self.assertEqual(0, self._events["event_always"])
        self.assertEqual(0, self._events["event_if_modeactive"])
        self.assertEqual(0, self._events["event_if_modestopping"])
        self.post_event("test_conditional_handlers")
        self.assertEqual(1, self._events["event_always"])
        self.assertEqual(0, self._events["event_if_modeactive"])
        self.assertEqual(0, self._events["event_if_modestopping"])

        # test one condition passing
        self.mock_event("event_always")
        self.mock_event("event_if_modeactive")
        self.mock_event("event_if_modestopping")

        self.machine.modes["mode1"].start()
        self.advance_time_and_run()
        self.assertEqual(0, self._events["event_always"])
        self.assertEqual(0, self._events["event_if_modeactive"])
        self.assertEqual(0, self._events["event_if_modestopping"])
        self.post_event("test_conditional_handlers")
        self.assertEqual(1, self._events["event_always"])
        self.assertEqual(1, self._events["event_if_modeactive"])
        self.assertEqual(0, self._events["event_if_modestopping"])

        # test both conditions passing
        self.mock_event("event_always")
        self.mock_event("event_if_modeactive")
        self.mock_event("event_if_modestopping")

        self.machine.modes["mode1"].stop()
        self.assertEqual(0, self._events["event_always"])
        self.assertEqual(0, self._events["event_if_modeactive"])
        self.assertEqual(0, self._events["event_if_modestopping"])
        self.post_event("test_conditional_handlers")
        self.assertEqual(1, self._events["event_always"])
        self.assertEqual(1, self._events["event_if_modeactive"])
        self.assertEqual(1, self._events["event_if_modestopping"])

    def test_event_time_delays(self):
        self.mock_event('td1')
        self.mock_event('td2')

        self.post_event('test_time_delay1')
        self.advance_time_and_run(1)
        self.assertEventNotCalled('td1')
        self.advance_time_and_run(1)
        self.assertEventCalled('td1')

        self.post_event('test_time_delay2')
        self.advance_time_and_run(1)
        self.assertEventNotCalled('td2')
        self.advance_time_and_run(1)
        self.assertEventCalled('td2')

    def test_mode_condition(self):
        self.mock_event('mode1_active')
        self.mock_event('mode1_not_active')

        self.assertFalse(self.machine.modes["mode1"].active)

        self.post_event('test_conditional_mode')

        self.assertEventNotCalled('mode1_active')
        self.assertEventCalled('mode1_not_active')

        self.mock_event('mode1_active')
        self.mock_event('mode1_not_active')

        self.machine.modes["mode1"].start()
        self.advance_time_and_run()

        self.post_event('test_conditional_mode')
        self.assertTrue(self.machine.modes["mode1"].active)

        self.assertEventCalled('mode1_active')
        self.assertEventNotCalled('mode1_not_active')

    def test_event_placeholder(self):
        self.mock_event('my_event_None_123')
        self.mock_event('my_event_hello_world_123')

        self.post_event("play_placeholder_event")
        self.assertEventCalled("my_event_None_123")
        self.mock_event('my_event_None_123')
        self.assertEventNotCalled("my_event_hello_world_123")

        self.machine.variables.set_machine_var("test", "hello_world")
        self.post_event("play_placeholder_event")
        self.assertEventNotCalled("my_event_None_123")
        self.assertEventCalled("my_event_hello_world_123")

    def test_arg_placeholder(self):
        self.mock_event('loaded_event_int')
        self.mock_event('loaded_event_float')
        self.mock_event('loaded_event_bool')
        self.mock_event('loaded_event_string')
        self.mock_event('loaded_event_notype')

        self.machine.variables.set_machine_var("testint", 1234)
        self.machine.variables.set_machine_var("testfloat", 12.34)
        self.machine.variables.set_machine_var("testbool", True)
        self.machine.variables.set_machine_var("teststring", "foobar")
        self.machine.variables.set_machine_var("testnotype", "barfoo")

        self.post_event("play_placeholder_args")
        self.assertEqual({"foo": 1234, "priority": 0}, self._last_event_kwargs['loaded_event_int'])
        self.assertEqual({"foo": 12.34, "priority": 0}, self._last_event_kwargs['loaded_event_float'])
        self.assertEqual({"foo": True, "priority": 0}, self._last_event_kwargs['loaded_event_bool'])
        self.assertEqual({"foo": "foobar", "priority": 0}, self._last_event_kwargs['loaded_event_string'])
        self.assertEqual({"foo": "barfoo", "priority": 0}, self._last_event_kwargs['loaded_event_notype'])

    def test_kwarg_placeholder(self):
        self.mock_event('event_always')
        self.mock_event('event_foobar')
        self.mock_event('event_(name)')

        self.post_event_with_params("play_event_with_kwargs", name="foobar")
        self.assertEventCalled("event_always")
        self.assertEventCalled("event_foobar")
        self.assertEventNotCalled("event_(name)")

    def test_value_kwarg_evaluation(self):
        self.mock_event('event_with_param_kwargs')

        self.post_event_with_params("play_event_with_param_kwargs", result="bar", initial=6)
        self.assertEqual({"foo": "bar", "maths": 30, "priority": 0}, self._last_event_kwargs["event_with_param_kwargs"])




"""A request was made to drive the driver outside its configured limits."""


class DriverLimitsError(AssertionError):

    """A request was made to drive the driver outside its configured limits."""



from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallDeviceAutoManualPlunger(MpfTestCase):

    def get_config_file(self):
        return 'test_ball_device_auto_manual_plunger.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def test_mechanical_eject_to_pf(self):

        plunger = self.machine.ball_devices['plunger']
        playfield = self.machine.ball_devices['playfield']
        trough_coil = self.machine.coils['trough_eject']
        plunger_coil = self.machine.coils['plunger_eject']
        trough_coil.pulse = MagicMock()
        plunger_coil.pulse = MagicMock()

        # add ball to trough
        self.machine.switch_controller.process_switch('s_trough_1', 1)

        self.advance_time_and_run(1)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, plunger.balls)
        assert not plunger_coil.pulse.called
        assert not trough_coil.pulse.called

        # Request player eject to pf
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        self.assertTrue(trough_coil.pulse.called)
        assert not plunger_coil.pulse.called

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.advance_time_and_run(1)

        # plunger gets ball
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, plunger.balls)
        assert not plunger_coil.pulse.called

        # player waits, then mechanically plunges
        self.advance_time_and_run(5)
        self.machine.switch_controller.process_switch('s_plunger', 0)

        # ball lands on pf
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.machine.switch_controller.process_switch('s_playfield', 0)
        self.advance_time_and_run(.1)

        assert not plunger_coil.pulse.called
        self.assertEqual(0, plunger.balls)
        self.assertEqual(1, playfield.balls)

    def test_powered_eject_to_pf(self):

        plunger = self.machine.ball_devices['plunger']
        playfield = self.machine.ball_devices['playfield']
        trough_coil = self.machine.coils['trough_eject']
        plunger_coil = self.machine.coils['plunger_eject']
        trough_coil.pulse = MagicMock()
        plunger_coil.pulse = MagicMock()

        # add ball to trough
        self.machine.switch_controller.process_switch('s_trough_1', 1)

        self.advance_time_and_run(1)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, plunger.balls)
        assert not plunger_coil.pulse.called
        assert not trough_coil.pulse.called

        # Request player eject to pf
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        self.assertTrue(trough_coil.pulse.called)
        assert not plunger_coil.pulse.called

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.advance_time_and_run(1)

        # plunger gets ball
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, plunger.balls)
        assert not plunger_coil.pulse.called

        # player waits, then hits launch button
        self.advance_time_and_run(5)
        self.machine.switch_controller.process_switch('s_launch', 1)
        self.machine.switch_controller.process_switch('s_launch', 0)

        self.advance_time_and_run(1)
        self.assertTrue(plunger_coil.pulse.called)

        self.machine.switch_controller.process_switch('s_plunger', 0)

        # ball lands on pf
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch('s_playfield', 1)
        self.machine.switch_controller.process_switch('s_playfield', 0)
        self.advance_time_and_run(.1)

        self.assertEqual(0, plunger.balls)
        self.assertEqual(1, playfield.balls)

    def test_returned_ball_with_missing_ball(self):
        plunger = self.machine.ball_devices['plunger']
        trough = self.machine.ball_devices['trough']
        playfield = self.machine.ball_devices['playfield']
        trough_coil = self.machine.coils['trough_eject']
        plunger_coil = self.machine.coils['plunger_eject']
        trough_coil.pulse = MagicMock()
        plunger_coil.pulse = MagicMock()

        # add ball to trough
        self.machine.switch_controller.process_switch('s_trough_1', 1)
        self.machine.switch_controller.process_switch('s_trough_2', 1)

        self.advance_time_and_run(1)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, plunger.balls)
        assert not plunger_coil.pulse.called
        assert not trough_coil.pulse.called

        # Request player eject to pf
        playfield.add_ball(player_controlled=False)
        self.advance_time_and_run(1)

        self.assertTrue(trough_coil.pulse.called)
        assert not plunger_coil.pulse.called

        self.machine.switch_controller.process_switch('s_trough_1', 0)
        self.advance_time_and_run(1)

        # plunger gets ball
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, plunger.balls)
        self.assertTrue(plunger_coil.pulse.called)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch('s_plunger', 0)
        self.advance_time_and_run(1)

        # ball returns
        self.machine.switch_controller.process_switch('s_plunger', 1)
        self.advance_time_and_run(9)
        # but then disappears
        self.machine.switch_controller.process_switch('s_plunger', 0)
        self.advance_time_and_run(20)

        self.assertEqual(0, plunger.balls)
        self.assertFalse(plunger.outgoing_balls_handler._eject_queue.qsize())
        self.assertEqual("idle", plunger.state)
        self.assertEqual(1, playfield.balls)
        self.assertFalse(trough.outgoing_balls_handler._eject_queue.qsize())
        self.assertEqual("idle", trough.state)
        self.assertBallsOnPlayfield(1)



"""VPE simulator.

This needs to be in a separate file so that we can handle syntax errors for python <= 3.5.
"""

import asyncio
from mpf.platforms.visual_pinball_engine import platform_pb2


class VpeSimulation:

    def __init__(self, switches):
        self.switches = switches
        self.change_queue = None
        self.lights = {}
        self.coils = {}
        self.rules = {}
        self.dmd_frames = {}

    def init_async(self):
        self.change_queue = asyncio.Queue()

    async def connect(self, service):
        configuration = platform_pb2.MachineState(
            initial_switch_states=self.switches)
        command_stream = service.Start(configuration, None)
        asyncio.ensure_future(self.read_commands(command_stream))

        asyncio.ensure_future(service.SendSwitchChanges(self.switch_changes(), None))

    async def switch_changes(self):
        while True:
            changed_switch = await self.change_queue.get()
            yield platform_pb2.SwitchChanges(switch_number=changed_switch[0], switch_state=changed_switch[1])

    async def read_commands(self, stream):
        async for command in stream:
            variant = command.WhichOneof("command")
            if variant == "fade_light":
                self.handle_fade_light(command.fade_light)
            elif variant == "pulse_coil":
                self.handle_pulse(command.pulse_coil)
            elif variant == "enable_coil":
                self.handle_enable(command.enable_coil)
            elif variant == "disable_coil":
                self.handle_disable(command.disable_coil)
            elif variant == "disable_coil":
                self.handle_disable(command.disable_coil)
            elif variant == "configure_hardware_rule":
                self.handle_rule(command.configure_hardware_rule)
            elif variant == "remove_hardware_rule":
                self.handle_rule_remove(command.remove_hardware_rule)
            elif variant == "dmd_frame_request":
                self.handle_dmd_frame_request(command.dmd_frame_request)
            else:
                raise AssertionError("Not implemented {}".format(variant))

    def set_switch(self, switch, state):
        self.switches[switch] = state
        self.change_queue.put_nowait((switch, state))

    def handle_fade_light(self, request):
        for fade in request.fades:
            if request.common_fade_ms > 0:
                self.lights[fade.light_number] = (asyncio.get_event_loop().time() +
                                                  (request.common_fade_ms / 1000.0), fade.target_brightness)
            else:
                self.lights[fade.light_number] = fade.target_brightness

    def handle_pulse(self, request):
        self.coils[request.coil_number] = "pulsed-{}-{}".format(request.pulse_ms, request.pulse_power)

    def handle_enable(self, request):
        self.coils[request.coil_number] = "enabled-{}-{}-{}".format(request.pulse_ms, request.pulse_power, request.hold_power)

    def handle_disable(self, request):
        self.coils[request.coil_number] = "disabled"

    def handle_rule(self, request):
        self.rules["{}-{}".format(request.coil_number, request.switch_number)] = request

    def handle_rule_remove(self, request):
        del self.rules["{}-{}".format(request.coil_number, request.switch_number)]

    def handle_dmd_frame_request(self, request):
        self.dmd_frames[request.name] = (request.frame, request.brightness)



from mpf.tests.MpfTestCase import MpfTestCase
from mpf._version import log_url


class TestModesConfigValidation(MpfTestCase):

    def get_config_file(self):
        return self.config

    def get_machine_path(self):
        return 'tests/machine_files/mode_tests/'

    def setUp(self):
        self.save_and_prepare_sys_path()

    def tearDown(self):
        self.restore_sys_path()

    def test_loading_invalid_modes(self):
        self.config = 'test_loading_invalid_modes.yaml'
        with self.assertRaises(AssertionError) as context:
            super(TestModesConfigValidation, self).setUp()

        self.loop.close()

        self.assertEqual("No config found for mode 'invalid'. MPF expects the config at "
                         "'modes/invalid/config/invalid.yaml' inside your machine folder.",
                         str(context.exception))

    def test_empty_modes_section(self):
        self.config = 'test_empty_modes_section.yaml'
        super(TestModesConfigValidation, self).setUp()
        super().tearDown()

    def test_broken_mode_config(self):
        self.config = 'test_broken_mode_config.yaml'
        with self.assertRaises(AssertionError) as context:
            super(TestModesConfigValidation, self).setUp()

        self.loop.close()

        self.maxDiff = None
        self.assertEqual('Config File Error in ConfigValidator: Your config contains a value for the setting '
                         '"mode:invalid_key", but this is not a valid setting name. Error Code: CFE-ConfigValidator-2 '
                         '({})'.format(log_url.format("CFE-ConfigValidator-2")),
                         str(context.exception))

    def test_missing_mode_section(self):
        self.config = 'test_missing_mode_section.yaml'
        super(TestModesConfigValidation, self).setUp()

        self.assertTrue("broken_mode2" in self.machine.modes)

        super().tearDown()

    def test_mode_without_config(self):
        self.config = 'test_mode_without_config.yaml'
        with self.assertRaises(AssertionError) as context:
            super(TestModesConfigValidation, self).setUp()

        self.loop.close()

        self.assertEqual("No config found for mode 'mode_without_config'. MPF expects the config at "
                         "'modes/mode_without_config/config/mode_without_config.yaml' inside your machine folder.",
                         str(context.exception))



from unittest.mock import MagicMock

from mpf.tests.MpfTestCase import MpfTestCase, test_config


class TestMotors(MpfTestCase):

    def get_machine_path(self):
        return 'tests/machine_files/motor/'

    @test_config("drop_target.yaml")
    def testMotorizedDropTargetBank(self):
        motor = self.machine.motors["motorized_drop_target_bank"]
        coil = self.machine.digital_outputs["c_motor_run"]
        coil.enable = MagicMock()
        coil.disable = MagicMock()

        # reset should move it down
        motor.reset()
        self.advance_time_and_run()
        coil.enable.assert_called_with()
        coil.enable = MagicMock()
        assert not coil.disable.called

        # it goes up. nothing should happen
        self.hit_switch_and_run("s_position_up", 1)
        assert not coil.enable.called
        assert not coil.disable.called

        # it leaves up position
        self.release_switch_and_run("s_position_up", 1)
        assert not coil.enable.called
        assert not coil.disable.called

        self.advance_time_and_run(5)
        # it goes down. motor should stop
        self.hit_switch_and_run("s_position_down", 1)
        assert not coil.enable.called
        coil.disable.assert_called_with()
        coil.disable = MagicMock()

        # should not start motor
        self.post_event("go_down2")
        self.advance_time_and_run()
        assert not coil.enable.called
        coil.disable.assert_called_with()
        coil.disable = MagicMock()

        # go up
        self.post_event("go_up")
        self.advance_time_and_run()
        coil.enable.assert_called_with()
        coil.enable = MagicMock()
        assert not coil.disable.called

        # it leaves down position
        self.release_switch_and_run("s_position_down", 0)
        assert not coil.enable.called
        assert not coil.disable.called

        self.advance_time_and_run(5)
        # it goes up. motor should stop
        self.hit_switch_and_run("s_position_up", 1)
        assert not coil.enable.called
        coil.disable.assert_called_with()
        coil.disable = MagicMock()

    @test_config("ghostbusters.yaml")
    def testMotorToyWithTwoEndSwitches(self):
        slimer = self.machine.motors["ghostbusters_slimer"]
        motor_left = self.machine.digital_outputs["c_slimer_motor_forward"].hw_driver
        motor_right = self.machine.digital_outputs["c_slimer_motor_backward"].hw_driver
        self.assertPlaceholderEvaluates("left", "device.motors.ghostbusters_slimer.move_direction")
        self.assertPlaceholderEvaluates(None, "device.motors.ghostbusters_slimer.last_position")
        self.assertPlaceholderEvaluates("home", "device.motors.ghostbusters_slimer.target_position")

        self.assertEqual(1.0, motor_left.current_brightness)
        self.assertEqual(0.0, motor_right.current_brightness)

        self.hit_switch_and_run("s_slimer_home", 1)
        self.assertEqual(0.0, motor_left.current_brightness)
        self.assertEqual(0.0, motor_right.current_brightness)
        self.assertPlaceholderEvaluates("stopped", "device.motors.ghostbusters_slimer.move_direction")
        self.assertPlaceholderEvaluates("home", "device.motors.ghostbusters_slimer.last_position")
        self.assertPlaceholderEvaluates("home", "device.motors.ghostbusters_slimer.target_position")

        slimer.go_to_position("away")
        self.advance_time_and_run()
        self.assertEqual(0.0, motor_left.current_brightness)
        self.assertEqual(1.0, motor_right.current_brightness)
        self.release_switch_and_run("s_slimer_home", 1)
        self.hit_switch_and_run("s_slimer_away", 1)

        self.assertEqual(0.0, motor_left.current_brightness)
        self.assertEqual(0.0, motor_right.current_brightness)

        slimer.go_to_position("away")
        self.assertEqual(0.0, motor_left.current_brightness)
        self.assertEqual(0.0, motor_right.current_brightness)

        slimer.go_to_position("home")
        self.assertEqual(1.0, motor_left.current_brightness)
        self.assertEqual(0.0, motor_right.current_brightness)

        self.hit_switch_and_run("s_slimer_home", 1)
        self.assertEqual(0.0, motor_left.current_brightness)
        self.assertEqual(0.0, motor_right.current_brightness)

    @test_config("multiposition_motor.yaml")
    def testMotorWithMultipleSwitches(self):
        """Test motor with home switch on the right."""
        self.assertPlaceholderEvaluates("right", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates(None, "device.motors.multiposition_motor.last_position")

        self.hit_switch_and_run("s_multiposition_motor_3", 1)
        self.assertPlaceholderEvaluates("right", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates("position3", "device.motors.multiposition_motor.last_position")

        self.release_switch_and_run("s_multiposition_motor_3", 1)
        self.hit_switch_and_run("s_multiposition_motor_4", 1)
        self.assertPlaceholderEvaluates("stopped", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates("position4", "device.motors.multiposition_motor.last_position")

    @test_config("multiposition_motor.yaml")
    def testMotorHitsUnexpectedEndSwitch(self):
        """Test if motor stops when it reached an unexpected end switch."""
        self.assertPlaceholderEvaluates("right", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates(None, "device.motors.multiposition_motor.last_position")

        # for some reason the motor reached the left end (while moving right)
        self.hit_switch_and_run("s_multiposition_motor_1", 1)
        self.assertPlaceholderEvaluates("stopped", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates("position1", "device.motors.multiposition_motor.last_position")

        # move motor again
        self.machine.motors["multiposition_motor"].go_to_position("position3")
        self.advance_time_and_run(1)
        # it should move right as it is on the left stop
        self.assertPlaceholderEvaluates("right", "device.motors.multiposition_motor.move_direction")

        self.release_switch_and_run("s_multiposition_motor_1", 1)

        # for some reason it misses position3 and bumps in to the end stop at the right
        self.hit_switch_and_run("s_multiposition_motor_4", 1)
        self.assertPlaceholderEvaluates("stopped", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates("position4", "device.motors.multiposition_motor.last_position")

    @test_config("multiposition_motor_start_on_end_switch.yaml")
    def testMotorOnEndSwitchAtStart(self):
        """Motor is at the right end switch but that is not home."""
        # motor should move left right away
        self.assertPlaceholderEvaluates("left", "device.motors.multiposition_motor.move_direction")

    @test_config("multiposition_motor.yaml")
    def testStuckSwitches(self):
        """Motor should stop when multiple position switches are active."""
        self.assertPlaceholderEvaluates("right", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates(None, "device.motors.multiposition_motor.last_position")

        # two switches are stuck - motor should stop
        self.hit_switch_and_run("s_multiposition_motor_2", 1)
        self.hit_switch_and_run("s_multiposition_motor_3", 1)
        self.assertPlaceholderEvaluates("stopped", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates(None, "device.motors.multiposition_motor.last_position")

        # motor should not move
        self.machine.motors["multiposition_motor"].go_to_position("position1")
        self.advance_time_and_run(1)
        self.assertPlaceholderEvaluates("stopped", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates(None, "device.motors.multiposition_motor.last_position")

        # switch 2 becomes unstuck
        self.release_switch_and_run("s_multiposition_motor_2", 1)

        # motor should move again
        self.machine.motors["multiposition_motor"].go_to_position("position2")
        self.advance_time_and_run(1)
        self.assertPlaceholderEvaluates("left", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates(None, "device.motors.multiposition_motor.last_position")

        self.release_switch_and_run("s_multiposition_motor_3", 1)
        self.hit_switch_and_run("s_multiposition_motor_2", 1)
        self.assertPlaceholderEvaluates("stopped", "device.motors.multiposition_motor.move_direction")
        self.assertPlaceholderEvaluates("position2", "device.motors.multiposition_motor.last_position")

    @test_config("multiposition_motor_home_in_the_middle.yaml")
    def testMotorWithMultipleSwitchesAndHomeInTheMiddle(self):
        """Motor starts moving to the right but should reverse when it knows its position."""
        self.assertPlaceholderEvaluates("right", "device.motors.multiposition_motor2.move_direction")
        self.assertPlaceholderEvaluates(None, "device.motors.multiposition_motor2.last_position")

        # first known position - motor reverses
        self.hit_switch_and_run("s_multiposition_motor_3", 1)
        self.assertPlaceholderEvaluates("left", "device.motors.multiposition_motor2.move_direction")
        self.assertPlaceholderEvaluates("position3", "device.motors.multiposition_motor2.last_position")

        # motor reaches home
        self.release_switch_and_run("s_multiposition_motor_3", 1)
        self.hit_switch_and_run("s_multiposition_motor_2", 1)
        self.assertPlaceholderEvaluates("stopped", "device.motors.multiposition_motor2.move_direction")
        self.assertPlaceholderEvaluates("position2", "device.motors.multiposition_motor2.last_position")



"""Standard pulse ejector."""
from typing import List

import asyncio

from mpf.devices.ball_device.ball_device_ejector import BallDeviceEjector
from mpf.devices.ball_device.default_ball_search import DefaultBallSearch
from mpf.exceptions.config_file_error import ConfigFileError


class PulseCoilEjector(DefaultBallSearch, BallDeviceEjector):

    """Pulse a coil to eject one ball."""

    __slots__ = []  # type: List[str]

    def __init__(self, config, ball_device, machine):
        """Initialise pulse coil ejector."""
        for option in ["eject_coil", "eject_coil_jam_pulse", "eject_coil_retry_pulse", "eject_coil_reorder_pulse",
                       "eject_coil_max_wait_ms", "retries_before_increasing_pulse"]:
            if option not in config and option in ball_device.config:
                config[option] = ball_device.config[option]

        super().__init__(config, ball_device, machine)

        self.config = self.machine.config_validator.validate_config("ball_device_ejector_pulse", self.config)

        # prevent conflicting options
        if "eject_coil_enable_time" in self.ball_device.config and self.ball_device.config['eject_coil_enable_time']:
            raise ConfigFileError("Pulse Coil Ejector does not support eject_coil_enable_time.", 2,
                                  self.ball_device.log.name + "-pulse_ejector")

    @staticmethod
    def _get_pulse_setting_for_ball_count(pulse_setting, balls_in_device):
        """Select the setting depending on the ball count in the device."""
        if len(pulse_setting) >= balls_in_device:
            return pulse_setting[balls_in_device - 1]

        # default -> last setting
        return pulse_setting[-1]

    async def eject_one_ball(self, is_jammed, eject_try, balls_in_device):
        """Pulse eject coil."""
        max_wait_ms = self.config['eject_coil_max_wait_ms']
        if eject_try <= 2 and self.config['eject_coil_jam_pulse'] and is_jammed:
            # jammed -> different pulse settings to only eject one ball
            self.config['eject_coil'].pulse(
                self._get_pulse_setting_for_ball_count(self.config['eject_coil_jam_pulse'], balls_in_device),
                max_wait_ms=max_wait_ms)
        elif eject_try >= self.config['retries_before_increasing_pulse'] and self.config['eject_coil_retry_pulse']:
            # multiple failed ejects -> increase pulse strength
            self.config['eject_coil'].pulse(
                self._get_pulse_setting_for_ball_count(self.config['eject_coil_retry_pulse'], balls_in_device),
                max_wait_ms=max_wait_ms)
        elif self.config["eject_times"]:
            # use pulse_times parameter
            self.config['eject_coil'].pulse(
                self._get_pulse_setting_for_ball_count(self.config['eject_times'], balls_in_device),
                max_wait_ms=max_wait_ms)
        else:
            # default pulse
            self.config['eject_coil'].pulse(max_wait_ms=max_wait_ms)

        self.ball_device.debug_log("Firing eject coil. Current balls: %s.", self.ball_device.balls)

    async def reorder_balls(self):
        """Reorder balls without ejecting."""
        if not self.config['eject_coil_reorder_pulse']:
            self.ball_device.log.warning("Cannot reorder device because eject_coil_reorder_pulse is not configured")
            return

        max_wait_ms = self.config['eject_coil_max_wait_ms']
        wait_ms = self.config['eject_coil'].pulse(self.config['eject_coil_reorder_pulse'], max_wait_ms=max_wait_ms)

        # wait for wait_ms + pulse_ms + 2s for sanity
        duration = wait_ms / 1000.0 + self.config['eject_coil_reorder_pulse'] / 1000.0 + 2.0
        await asyncio.sleep(duration)

    def _fire_coil_for_search(self, full_power):
        if not full_power and self.config['eject_coil_jam_pulse']:
            self.config['eject_coil'].pulse(self.config['eject_coil_jam_pulse'][0])
        else:
            self.config['eject_coil'].pulse()
        return True



import os

import argparse
import re
import unittest
import sys

# some hacks to unbreak logging afte loading the kivy logger
import logging
root = logging.root
os.environ['KIVY_NO_FILELOG'] = '1'
os.environ['KIVY_NO_CONSOLELOG'] = '1'
os.environ["KIVY_NO_ARGS"] = "1"

from kivy.logger import Logger
for handler in Logger.handlers:
    Logger.removeHandler(handler)
sys.stdout = sys.__stdout__
logging.root = root
# end of anti kivy hacks

from mpf.commands import MpfCommandLineParser
from mpf.tests.MpfDocTestCase import MpfDocTestCase, MpfDocTestCaseNoFakeGame
from mpf.tests.MpfIntegrationDocTestCase import MpfIntegrationDocTestCase

SUBCOMMAND = True


class Command(MpfCommandLineParser):

    """Run a text unit test from cli."""

    def __init__(self, args, path):
        """Parse args."""
        super().__init__(args, path)
        test_file = self.argv.pop(1)

        parser = argparse.ArgumentParser(description='MPF Command')

        parser.add_argument("-v", help="verbose",
                            default=False, action="store_true", dest="verbose")
        parser.add_argument("-m", help="Start MC on non sphinx test (auto-detected in sphinx)",
                            default=False, action="store_true", dest="start_mc")

        args = parser.parse_args(self.argv[1:])

        with open(test_file) as f:
            test_string = f.read()

        base_dir = os.path.dirname(os.path.abspath(test_file))
        suite = unittest.TestSuite()
        simulation = "##! no_simulation" not in test_string

        if ".. code-block:: mpf-config" in test_string or ".. code-block:: mpf-mc-config" in test_string:
            print("Parsing documentation page")
            blocks = re.finditer(".. code-block:: (?P<type>mpf-config|mpf-mc-config)\n\n(?P<code>( {2,4}[^\n]+|\n)+)",
                                 test_string)
            if not blocks:
                raise AssertionError("Could not parse tests.")

            for num, block in enumerate(blocks):
                text = block.group("code")
                indent_len = None
                test_case = ''
                for line in text.split("\n"):
                    if indent_len is None:
                        indent = re.search("^( )+", line)
                        if not indent:
                            raise AssertionError("Block {} (starting at 0) is incorrectly indented.".format(num))
                        indent_len = len(indent.group(0))
                    test_case += line[indent_len:] + "\n"
                if block.group("type") == "mpf-config":
                    if "##! no_fake_game" in test_case:
                        test = MpfDocTestCaseNoFakeGame(config_string=test_case, base_dir=base_dir,
                                                        simulation=simulation)
                    else:
                        test = MpfDocTestCase(config_string=test_case, base_dir=base_dir, simulation=simulation)
                else:
                    test = MpfIntegrationDocTestCase(config_string=test_case, base_dir=base_dir, simulation=simulation)
                test._testMethodDoc = "Block {} (starting at 0)".format(num)
                suite.addTest(test)
        else:
            print("Parsing single test")

            if args.start_mc:
                test = MpfIntegrationDocTestCase(config_string=test_string, base_dir=base_dir, simulation=simulation)
            else:
                if "##! no_fake_game" in test_string:
                    test = MpfDocTestCaseNoFakeGame(config_string=test_string, base_dir=base_dir, simulation=simulation)
                else:
                    test = MpfDocTestCase(config_string=test_string, base_dir=base_dir, simulation=simulation)
            suite.addTest(test)

        result = unittest.TextTestRunner(verbosity=1 if not args.verbose else 99).run(suite)

        sys.exit(not result.wasSuccessful())



"""Fast servo implementation."""
from mpf.core.utility_functions import Util
from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface


class FastServo(ServoPlatformInterface):

    """A servo in the FAST platform."""

    __slots__ = ["number", "net_connection"]

    def __init__(self, number, net_connection):
        """Initialise servo."""
        self.number = number
        self.net_connection = net_connection

    def go_to_position(self, position):
        """Set a servo position."""
        if position < 0 or position > 1:
            raise AssertionError("Position has to be between 0 and 1")

        # convert from [0,1] to [0, 255]
        position_numeric = int(position * 255)

        cmd = 'XO:{},{}'.format(
            self.number,
            Util.int_to_hex_string(position_numeric))

        self.net_connection.send(cmd)

    def set_speed_limit(self, speed_limit):
        """Not implemented."""

    def set_acceleration_limit(self, acceleration_limit):
        """Not implemented."""
